import json
import os
import platform
import re
import textwrap

import pytest
from unittest import mock

from conan.tools.cmake.presets import load_cmake_presets
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.internal.util.files import rmdir, load


def test_cross_build():
    windows_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        """)
    rpi_profile = textwrap.dedent("""
        [settings]
        os=Linux
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        arch=armv8
        build_type=Release
        """)
    embedwin = textwrap.dedent("""
        [settings]
        os=Windows
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        arch=armv8
        build_type=Release
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                 "rpi": rpi_profile,
                 "embedwin": embedwin,
                 "windows": windows_profile})
    client.run("install . --profile:build=windows --profile:host=rpi")
    toolchain = client.load("conan_toolchain.cmake")

    assert "set(CMAKE_SYSTEM_NAME Linux)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR aarch64)" in toolchain
    # Revert of https://github.com/conan-io/conan/pull/18559, even if it was correct
    # assert "set(CMAKE_TRY_COMPILE_CONFIGURATION Release)" in toolchain

    client.run("install . --profile:build=windows --profile:host=embedwin")
    toolchain = client.load("conan_toolchain.cmake")

    assert "set(CMAKE_SYSTEM_NAME Windows)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR ARM64)" in toolchain


def test_cross_build_linux_to_macos():
    linux_profile = textwrap.dedent("""
        [settings]
        os=Linux
        arch=x86_64
        """)
    macos_profile = textwrap.dedent("""
        [settings]
        os=Macos
        os.version=13.1
        arch=x86_64
        compiler=apple-clang
        compiler.version=13
        compiler.libcxx=libc++
        build_type=Release
        """)

    client = TestClient(path_with_spaces=False)

    client.save({"conanfile.txt": "[generators]\nCMakeToolchain",
                 "linux": linux_profile,
                 "macos": macos_profile})
    client.run("install . --profile:build=linux --profile:host=macos")
    toolchain = client.load("conan_toolchain.cmake")

    assert "set(CMAKE_SYSTEM_NAME Darwin)" in toolchain
    assert "set(CMAKE_SYSTEM_VERSION 22)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR x86_64)" in toolchain
    # Revert of https://github.com/conan-io/conan/pull/18559, even if it was correct
    # assert "set(CMAKE_TRY_COMPILE_CONFIGURATION Release)" in toolchain


def test_cross_build_user_toolchain():
    # When a user_toolchain is defined in [conf], CMakeToolchain will not generate anything
    # for cross-build
    windows_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        """)
    rpi_profile = textwrap.dedent("""
        [settings]
        os=Linux
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        arch=armv8
        build_type=Release
        [conf]
        tools.cmake.cmaketoolchain:user_toolchain+=rpi_toolchain.cmake
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                "rpi": rpi_profile,
                 "windows": windows_profile})
    client.run("install . --profile:build=windows --profile:host=rpi")
    toolchain = client.load("conan_toolchain.cmake")

    # Fixed in https://github.com/conan-io/conan/issues/16807
    expected = textwrap.dedent("""\
        # Cross building
        if(NOT DEFINED CMAKE_SYSTEM_NAME) # It might have been defined by a user toolchain
        set(CMAKE_SYSTEM_NAME Linux)
        endif()
        if(NOT DEFINED CMAKE_SYSTEM_PROCESSOR) # It might have been defined by a user toolchain
        set(CMAKE_SYSTEM_PROCESSOR aarch64)
        endif()""")
    assert expected in toolchain


def test_cross_build_user_toolchain_confs():
    # When a user_toolchain is defined in [conf], but other confs are defined, they will be used
    windows_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        """)
    rpi_profile = textwrap.dedent("""
        [settings]
        os=Linux
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        arch=armv8
        build_type=Release
        [conf]
        tools.cmake.cmaketoolchain:user_toolchain+=rpi_toolchain.cmake
        tools.cmake.cmaketoolchain:system_name=Linux
        tools.cmake.cmaketoolchain:system_processor=aarch64
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                "rpi": rpi_profile,
                 "windows": windows_profile})
    client.run("install . --profile:build=windows --profile:host=rpi")
    toolchain = client.load("conan_toolchain.cmake")

    assert "set(CMAKE_SYSTEM_NAME Linux)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR aarch64)" in toolchain


def test_no_cross_build():
    windows_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        build_type=Release
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                 "windows": windows_profile})
    client.run("install . --profile:build=windows --profile:host=windows")
    toolchain = client.load("conan_toolchain.cmake")

    assert "CMAKE_SYSTEM_NAME " not in toolchain
    assert "CMAKE_SYSTEM_PROCESSOR" not in toolchain


def test_cross_arch():
    # Compiling to 32bits in an architecture that runs is not full cross-compiling
    build_profile = textwrap.dedent("""
        [settings]
        os=Linux
        arch=x86_64
        """)
    profile_arm = textwrap.dedent("""
        [settings]
        os=Linux
        arch=armv8
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        build_type=Release
        """)
    profile_macos = textwrap.dedent("""
        [settings]
        os=Macos
        arch=armv8
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        build_type=Release
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                 "linux64": build_profile,
                 "macos": profile_macos,
                 "linuxarm": profile_arm})
    client.run("install . --profile:build=linux64 --profile:host=linuxarm")
    toolchain = client.load("conan_toolchain.cmake")

    assert "set(CMAKE_SYSTEM_NAME Linux)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR aarch64)" in toolchain

    client.run("install . --profile:build=linux64 --profile:host=macos")
    toolchain = client.load("conan_toolchain.cmake")

    assert "set(CMAKE_SYSTEM_NAME Darwin)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR arm64)" in toolchain


def test_no_cross_build_arch():
    # Compiling to 32bits in an architecture that runs is not full cross-compiling
    build_profile = textwrap.dedent("""
        [settings]
        os=Linux
        arch=x86_64
        """)
    profile_32 = textwrap.dedent("""
        [settings]
        os=Linux
        arch=x86
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        build_type=Release
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                "linux64": build_profile,
                 "linux32": profile_32})
    client.run("install . --profile:build=linux64 --profile:host=linux32")
    toolchain = client.load("conan_toolchain.cmake")

    assert "CMAKE_SYSTEM_NAME" not in toolchain
    assert "CMAKE_SYSTEM_PROCESSOR " not in toolchain


def test_cross_build_conf():
    windows_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        """)
    rpi_profile = textwrap.dedent("""
        [settings]
        os=Linux
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        arch=armv8
        build_type=Release

        [conf]
        tools.cmake.cmaketoolchain:system_name = Custom
        tools.cmake.cmaketoolchain:system_version= 42
        tools.cmake.cmaketoolchain:system_processor = myarm
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                "rpi": rpi_profile,
                 "windows": windows_profile})
    client.run("install . --profile:build=windows --profile:host=rpi")
    toolchain = client.load("conan_toolchain.cmake")

    assert "set(CMAKE_SYSTEM_NAME Custom)" in toolchain
    assert "set(CMAKE_SYSTEM_VERSION 42)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR myarm)" in toolchain


def test_find_builddirs():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Conan(ConanFile):

            def package_info(self):
                self.cpp_info.builddirs = ["/path/to/builddir"]
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=dep --version=1.0")

    conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.cmake import CMakeToolchain

            class Conan(ConanFile):
                settings = "os", "arch", "compiler", "build_type"
                requires = "dep/1.0@"

                def generate(self):
                    cmake = CMakeToolchain(self)
                    cmake.generate()
            """)

    client.save({"conanfile.py": conanfile})
    client.run("install . ")
    with open(os.path.join(client.current_folder, "conan_toolchain.cmake")) as f:
        contents = f.read()
        assert "/path/to/builddir" in contents

    conanfile = textwrap.dedent("""
       import os
       from conan import ConanFile
       from conan.tools.cmake import CMakeToolchain

       class Conan(ConanFile):
           name = "mydep"
           version = "1.0"
           settings = "os", "arch", "compiler", "build_type"

           def build_requirements(self):
               self.test_requires("dep/1.0")

           def generate(self):
               cmake = CMakeToolchain(self)
               cmake.generate()
       """)

    client.save({"conanfile.py": conanfile})
    client.run("install . ")
    with open(os.path.join(client.current_folder, "conan_toolchain.cmake")) as f:
        contents = f.read()
        assert "/path/to/builddir" in contents


@pytest.fixture
def lib_dir_setup():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_generator("CMakeToolchain")})
    client.run("create . --name=onelib --version=1.0")
    client.run("create . --name=twolib --version=1.0")
    conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Conan(ConanFile):
                requires = "onelib/1.0", "twolib/1.0"

            """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=dep --version=1.0")

    conanfile = (GenConanfile().with_requires("dep/1.0").with_generator("CMakeToolchain")
                 .with_settings("os", "arch", "compiler", "build_type"))

    client.save({"conanfile.py": conanfile})
    return client


def test_runtime_lib_dirs_single_conf(lib_dir_setup):
    client = lib_dir_setup
    generator = ""
    is_windows = platform.system() == "Windows"
    if is_windows:
        generator = '-c tools.cmake.cmaketoolchain:generator=Ninja'

    client.run(f'install . -s build_type=Release {generator}')
    contents = client.load("conan_toolchain.cmake")
    pattern_lib_path = r'list\(PREPEND CMAKE_LIBRARY_PATH (.*)\)'
    pattern_lib_dirs = r'set\(CONAN_RUNTIME_LIB_DIRS (.*) \)'

    # On *nix platforms: the list in `CMAKE_LIBRARY_PATH`
    # is the same as `CONAN_RUNTIME_LIB_DIRS`
    # On windows, it's the same but with `bin` instead of `lib`
    cmake_library_path = re.search(pattern_lib_path, contents).group(1)
    conan_runtime_lib_dirs = re.search(pattern_lib_dirs, contents).group(1)
    lib_path = cmake_library_path.replace("/p/lib", "/p/bin") if is_windows else cmake_library_path

    assert lib_path == conan_runtime_lib_dirs


def test_runtime_lib_dirs_multiconf(lib_dir_setup):
    client = lib_dir_setup
    generator = ""
    if platform.system() != "Windows":
        generator = '-c tools.cmake.cmaketoolchain:generator="Ninja Multi-Config"'

    client.run(f'install . -s build_type=Release {generator}')
    client.run(f'install . -s build_type=Debug {generator}')

    contents = client.load("conan_toolchain.cmake")
    pattern_lib_dirs = r"set\(CONAN_RUNTIME_LIB_DIRS ([^)]*)\)"
    runtime_lib_dirs = re.search(pattern_lib_dirs, contents).group(1)

    assert "<CONFIG:Release>" in runtime_lib_dirs
    assert "<CONFIG:Debug>" in runtime_lib_dirs


@pytest.mark.skipif(platform.system() != "Darwin", reason="Only OSX")
def test_cmaketoolchain_cmake_system_processor_cross_apple():
    """
    https://github.com/conan-io/conan/pull/10434
    CMAKE_SYSTEM_PROCESSOR was not set when cross-building in Mac
    """
    client = TestClient()
    client.save({"hello.py": GenConanfile().with_name("hello")
                                           .with_version("1.0")
                                           .with_settings("os", "arch", "compiler", "build_type")})
    profile_ios = textwrap.dedent("""
        include(default)
        [settings]
        os=iOS
        os.version=15.4
        os.sdk=iphoneos
        os.sdk_version=15.0
        arch=armv8
    """)
    client.save({"profile_ios": profile_ios})
    client.run("install hello.py -pr:h=./profile_ios -pr:b=default -g CMakeToolchain")
    toolchain = client.load("conan_toolchain.cmake")
    assert "set(CMAKE_SYSTEM_NAME iOS)" in toolchain
    assert "set(CMAKE_SYSTEM_VERSION 21)" in toolchain
    assert "set(CMAKE_SYSTEM_PROCESSOR arm64)" in toolchain


@pytest.mark.skipif(platform.system() != "Darwin", reason="Only OSX")
def test_apple_vars_overwrite_user_conf():
    """
        tools.cmake.cmaketoolchain:system_name and tools.cmake.cmaketoolchain:system_version
        will be overwritten by the apple block
    """
    client = TestClient()
    client.save({"hello.py": GenConanfile().with_name("hello")
                                           .with_version("1.0")
                                           .with_settings("os", "arch", "compiler", "build_type")})
    profile_ios = textwrap.dedent("""
        include(default)
        [settings]
        os=iOS
        os.version=15.4
        os.sdk=iphoneos
        os.sdk_version=15.0
        arch=armv8
    """)
    client.save({"profile_ios": profile_ios})
    client.run("install hello.py -pr:h=./profile_ios -pr:b=default -g CMakeToolchain "
               "-c tools.cmake.cmaketoolchain:system_name=tvOS "
               "-c tools.cmake.cmaketoolchain:system_version=15.1 "
               "-c tools.cmake.cmaketoolchain:system_processor=x86_64 ")

    toolchain = client.load("conan_toolchain.cmake")

    # should set the conf values but system/version are overwritten by the apple block
    assert "CMAKE_SYSTEM_NAME tvOS" in toolchain
    assert "CMAKE_SYSTEM_NAME iOS" not in toolchain
    assert "CMAKE_SYSTEM_VERSION 15.1" in toolchain
    assert "CMAKE_SYSTEM_VERSION 15.0" not in toolchain
    assert "CMAKE_SYSTEM_PROCESSOR x86_64" in toolchain
    assert "CMAKE_SYSTEM_PROCESSOR armv8" not in toolchain


def test_extra_flags_via_conf():
    profile = textwrap.dedent("""
        [settings]
        os=Linux
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        arch=armv8
        build_type=Release

        [conf]
        tools.build:cxxflags=["--flag1", "--flag2"]
        tools.build:cflags+=["--flag3", "--flag4"]
        tools.build:sharedlinkflags=+["--flag5", "--flag6"]
        tools.build:exelinkflags=["--flag7", "--flag8"]
        tools.build:defines=["D1", "D2"]
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                "profile": profile})
    client.run("install . --profile:build=profile --profile:host=profile")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'string(APPEND CONAN_CXX_FLAGS " --flag1 --flag2")' in toolchain
    assert 'string(APPEND CONAN_C_FLAGS " --flag3 --flag4")' in toolchain
    assert 'string(APPEND CONAN_SHARED_LINKER_FLAGS " --flag5 --flag6")' in toolchain
    assert 'string(APPEND CONAN_EXE_LINKER_FLAGS " --flag7 --flag8")' in toolchain
    assert 'add_compile_definitions( "D1" "D2")' in toolchain


def test_bitcode_enable_flag():
    profile = textwrap.dedent("""
        [settings]
        os=Macos
        arch=armv8
        compiler=apple-clang
        compiler.version=17
        compiler.libcxx=libc++
        build_type=Release

        [conf]
        tools.apple:enable_bitcode=True
        """)

    client = TestClient(path_with_spaces=False)

    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type") \
                              .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                 "profile": profile})
    client.run("install . --profile:build=profile --profile:host=profile")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'set(BITCODE "-fembed-bitcode")' in toolchain


def test_cmake_presets_binary_dir_available():
    client = TestClient(path_with_spaces=False)
    conanfile = textwrap.dedent("""
    from conan import ConanFile
    from conan.tools.cmake import cmake_layout
    class HelloConan(ConanFile):
        generators = "CMakeToolchain"
        settings = "os", "compiler", "build_type", "arch"

        def layout(self):
            cmake_layout(self)

    """)

    client.save({"conanfile.py": conanfile})
    client.run("install .")
    if platform.system() != "Windows":
        build_dir = os.path.join(client.current_folder, "build", "Release")
    else:
        build_dir = os.path.join(client.current_folder, "build")

    presets = load_cmake_presets(os.path.join(build_dir, "generators"))
    assert presets["configurePresets"][0]["binaryDir"] == build_dir


@pytest.mark.parametrize("presets", ["CMakePresets.json", "CMakeUserPresets.json"])
def test_cmake_presets_shared_preset(presets):
    """valid user preset file is created when multiple project presets inherit
    from the same conan presets.
    """
    client = TestClient()
    project_presets = textwrap.dedent("""
    {
        "version": 4,
        "cmakeMinimumRequired": {
            "major": 3,
            "minor": 23,
            "patch": 0
        },
        "include":["ConanPresets.json"],
        "configurePresets": [
            {
                "name": "debug1",
                "inherits": ["conan-debug"]
            },
            {
                "name": "debug2",
                "inherits": ["conan-debug"]
            },
            {
                "name": "release1",
                "inherits": ["conan-release"]
            },
            {
                "name": "release2",
                "inherits": ["conan-release"]
            }
        ]
    }""")
    conanfile = textwrap.dedent("""
    from conan import ConanFile
    from conan.tools.cmake import cmake_layout, CMakeToolchain

    class TestPresets(ConanFile):
        generators = ["CMakeDeps"]
        settings = "build_type"

        def layout(self):
            cmake_layout(self)

        def generate(self):
            tc = CMakeToolchain(self)
            tc.user_presets_path = 'ConanPresets.json'
            tc.generate()
    """)

    client.save({presets: project_presets,
                 "conanfile.py": conanfile,
                 "CMakeLists.txt": ""})  # File must exist for Conan to do Preset things.

    client.run("install . -s build_type=Debug")

    conan_presets = json.loads(client.load("ConanPresets.json"))
    assert len(conan_presets["configurePresets"]) == 1
    assert conan_presets["configurePresets"][0]["name"] == "conan-release"


def test_cmake_presets_multiconfig():
    client = TestClient()
    profile = textwrap.dedent("""
        [settings]
        os = Windows
        arch = x86_64
        compiler=msvc
        compiler.version=193
        compiler.runtime=static
        compiler.runtime_type=Release
    """)
    client.save({"conanfile.py": GenConanfile("mylib", "1.0"), "profile": profile})
    client.run("create . -s build_type=Release --profile:h=profile")
    client.run("create . -s build_type=Debug --profile:h=profile")

    client.run("install --requires=mylib/1.0@ -g CMakeToolchain "
               "-s build_type=Release --profile:h=profile")
    presets = json.loads(client.load("CMakePresets.json"))
    assert len(presets["configurePresets"]) == 1
    assert len(presets["buildPresets"]) == 1
    assert presets["buildPresets"][0]["configuration"] == "Release"
    assert len(presets["testPresets"]) == 1
    assert presets["testPresets"][0]["configuration"] == "Release"

    client.run("install --requires=mylib/1.0@ -g CMakeToolchain "
               "-s build_type=Debug --profile:h=profile")
    presets = json.loads(client.load("CMakePresets.json"))
    assert len(presets["configurePresets"]) == 1
    assert len(presets["buildPresets"]) == 2
    assert presets["buildPresets"][0]["configuration"] == "Release"
    assert presets["buildPresets"][1]["configuration"] == "Debug"
    assert len(presets["testPresets"]) == 2
    assert presets["testPresets"][0]["configuration"] == "Release"
    assert presets["testPresets"][1]["configuration"] == "Debug"

    client.run("install --requires=mylib/1.0@ -g CMakeToolchain "
               "-s build_type=RelWithDebInfo --profile:h=profile")
    client.run("install --requires=mylib/1.0@ -g CMakeToolchain "
               "-s build_type=MinSizeRel --profile:h=profile")

    presets = json.loads(client.load("CMakePresets.json"))
    assert len(presets["configurePresets"]) == 1
    assert len(presets["buildPresets"]) == 4
    assert presets["buildPresets"][0]["configuration"] == "Release"
    assert presets["buildPresets"][1]["configuration"] == "Debug"
    assert presets["buildPresets"][2]["configuration"] == "RelWithDebInfo"
    assert presets["buildPresets"][3]["configuration"] == "MinSizeRel"
    assert len(presets["testPresets"]) == 4
    assert presets["testPresets"][0]["configuration"] == "Release"
    assert presets["testPresets"][1]["configuration"] == "Debug"
    assert presets["testPresets"][2]["configuration"] == "RelWithDebInfo"
    assert presets["testPresets"][3]["configuration"] == "MinSizeRel"

    # Repeat one
    client.run("install --requires=mylib/1.0@ -g CMakeToolchain "
               "-s build_type=Debug --profile:h=profile")
    client.run("install --requires=mylib/1.0@ -g CMakeToolchain "
               "-s build_type=Debug --profile:h=profile")
    presets = json.loads(client.load("CMakePresets.json"))
    assert len(presets["configurePresets"]) == 1
    assert len(presets["buildPresets"]) == 4
    assert presets["buildPresets"][0]["configuration"] == "Release"
    assert presets["buildPresets"][1]["configuration"] == "Debug"
    assert presets["buildPresets"][2]["configuration"] == "RelWithDebInfo"
    assert presets["buildPresets"][3]["configuration"] == "MinSizeRel"

    assert len(presets["testPresets"]) == 4
    assert presets["testPresets"][0]["configuration"] == "Release"
    assert presets["testPresets"][1]["configuration"] == "Debug"
    assert presets["testPresets"][2]["configuration"] == "RelWithDebInfo"
    assert presets["testPresets"][3]["configuration"] == "MinSizeRel"

    assert len(presets["configurePresets"]) == 1
    assert presets["configurePresets"][0]["name"] == "conan-default"


def test_cmake_presets_singleconfig():
    """ without defining a layout, single config always overwrites
    the existing CMakePresets.json
    """
    client = TestClient()
    profile = textwrap.dedent("""
        [settings]
        os = Linux
        arch = x86_64
        compiler=gcc
        compiler.version=8
    """)
    client.save({"conanfile.py": GenConanfile("mylib", "1.0"), "profile": profile})
    client.run("create . -s build_type=Release --profile:h=profile")
    client.run("create . -s build_type=Debug --profile:h=profile")

    client.run("install --requires=mylib/1.0@ "
               "-g CMakeToolchain -s build_type=Release --profile:h=profile")
    presets = json.loads(client.load("CMakePresets.json"))
    assert len(presets["configurePresets"]) == 1
    assert presets["configurePresets"][0]["name"] == "conan-release"

    assert len(presets["buildPresets"]) == 1
    assert presets["buildPresets"][0]["configurePreset"] == "conan-release"

    assert len(presets["testPresets"]) == 1
    assert presets["testPresets"][0]["configurePreset"] == "conan-release"

    # This overwrites the existing profile, as there is no layout
    client.run("install --requires=mylib/1.0@ "
               "-g CMakeToolchain -s build_type=Debug --profile:h=profile")

    presets = json.loads(client.load("CMakePresets.json"))
    assert len(presets["configurePresets"]) == 1
    assert presets["configurePresets"][0]["name"] == "conan-debug"

    assert len(presets["buildPresets"]) == 1
    assert presets["buildPresets"][0]["configurePreset"] == "conan-debug"

    assert len(presets["testPresets"]) == 1
    assert presets["testPresets"][0]["configurePreset"] == "conan-debug"

    # Repeat configuration, it shouldn't add a new one
    client.run("install --requires=mylib/1.0@ "
               "-g CMakeToolchain -s build_type=Debug --profile:h=profile")
    presets = json.loads(client.load("CMakePresets.json"))
    assert len(presets["configurePresets"]) == 1


def test_toolchain_cache_variables():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain

        class Conan(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            options = {"enable_foobar": [True, False], "qux": ["ANY"], "number": [1,2]}
            default_options = {"enable_foobar": True, "qux": "baz", "number": 1}

            def generate(self):
                toolchain = CMakeToolchain(self)
                toolchain.cache_variables["foo"] = True
                toolchain.cache_variables["foo2"] = False
                toolchain.cache_variables["var"] = "23"
                toolchain.cache_variables["ENABLE_FOOBAR"] = self.options.enable_foobar
                toolchain.cache_variables["QUX"] = self.options.qux
                toolchain.cache_variables["NUMBER"] = self.options.number
                toolchain.cache_variables["CMAKE_SH"] = "THIS VALUE HAS PRIORITY"
                toolchain.cache_variables["CMAKE_POLICY_DEFAULT_CMP0091"] = "THIS VALUE HAS PRIORITY"
                toolchain.cache_variables["CMAKE_MAKE_PROGRAM"] = "THIS VALUE HAS NO PRIORITY"
                toolchain.generate()
        """)
    client.save({"conanfile.py": conanfile})
    with mock.patch("platform.system", mock.MagicMock(return_value="Windows")):
        client.run("install . --name=mylib --version=1.0 "
                   "-c tools.cmake.cmaketoolchain:generator='MinGW Makefiles' "
                   "-c tools.gnu:make_program='MyMake' -c tools.build:skip_test=True")

    presets = json.loads(client.load("CMakePresets.json"))
    cache_variables = presets["configurePresets"][0]["cacheVariables"]
    assert cache_variables["foo"] == 'ON'
    assert cache_variables["foo2"] == 'OFF'
    assert cache_variables["var"] == '23'
    assert cache_variables["CMAKE_SH"] == "THIS VALUE HAS PRIORITY"
    assert cache_variables["CMAKE_POLICY_DEFAULT_CMP0091"] == "THIS VALUE HAS PRIORITY"
    assert cache_variables["CMAKE_MAKE_PROGRAM"] == "MyMake"
    assert cache_variables["BUILD_TESTING"] == 'OFF'
    assert cache_variables["ENABLE_FOOBAR"] == 'ON'
    assert cache_variables["QUX"] == 'baz'
    assert cache_variables["NUMBER"] == 1

    def _format_val(val):
        return f'"{val}"' if isinstance(val, str) and " " in val else f"{val}"

    for var, value in cache_variables.items():
        assert f"-D{var}={_format_val(value)}" in client.out
    assert "-DCMAKE_TOOLCHAIN_FILE=" in client.out
    assert f"-G {_format_val('MinGW Makefiles')}" in client.out

    client.run("install . --name=mylib --version=1.0 -c tools.gnu:make_program='MyMake'")
    presets = json.loads(client.load("CMakePresets.json"))
    cache_variables = presets["configurePresets"][0]["cacheVariables"]
    assert cache_variables["CMAKE_MAKE_PROGRAM"] == "MyMake"


def test_variables_types():
    # https://github.com/conan-io/conan/pull/10941
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain

        class Conan(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            def generate(self):
                toolchain = CMakeToolchain(self)
                toolchain.variables["FOO"] = True
                toolchain.generate()
        """)
    client.save({"conanfile.py": conanfile})
    client.run("install . --name=mylib --version=1.0")

    toolchain = client.load("conan_toolchain.cmake")
    assert 'set(FOO ON CACHE BOOL "Variable FOO conan-toolchain defined")' in toolchain


def test_android_c_library():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Conan(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            generators = "CMakeToolchain"

            def configure(self):
                if self.settings.compiler != "msvc":
                    del self.settings.compiler.libcxx

        """)
    client.save({"conanfile.py": conanfile})
    # Settings
    settings = "-s arch=x86_64 -s os=Android -s os.api_level=23 -c tools.android:ndk_path=/foo"
    # Checking the Android variables created
    # Issue: https://github.com/conan-io/conan/issues/11798
    client.run("install . " + settings)
    conan_toolchain = client.load(os.path.join(client.current_folder, "conan_toolchain.cmake"))
    assert "set(ANDROID_PLATFORM android-23)" in conan_toolchain
    assert "set(ANDROID_ABI x86_64)" in conan_toolchain
    assert 'include("/foo/build/cmake/android.toolchain.cmake")' in conan_toolchain
    client.run("create . --name=foo --version=1.0 " + settings)


@pytest.mark.parametrize("cmake_legacy_toolchain", [True, False, None])
def test_android_legacy_toolchain_flag(cmake_legacy_toolchain):
    client = TestClient()
    conanfile = GenConanfile().with_settings("os", "arch")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile})
    settings = "-s arch=x86_64 -s os=Android -s os.api_level=23 -c tools.android:ndk_path=/foo"
    expected = None
    if cmake_legacy_toolchain is not None:
        settings += f" -c tools.android:cmake_legacy_toolchain={cmake_legacy_toolchain}"
        expected = "ON" if cmake_legacy_toolchain else "OFF"
    client.run("install . " + settings)
    conan_toolchain = client.load(os.path.join(client.current_folder, "conan_toolchain.cmake"))
    if cmake_legacy_toolchain is not None:
        assert f"set(ANDROID_USE_LEGACY_TOOLCHAIN_FILE {expected})" in conan_toolchain
    else:
        assert "ANDROID_USE_LEGACY_TOOLCHAIN_FILE" not in conan_toolchain


@pytest.mark.parametrize("cmake_legacy_toolchain", [True, False, None])
def test_android_legacy_toolchain_with_compileflags(cmake_legacy_toolchain):
    # https://github.com/conan-io/conan/issues/13374
    client = TestClient()
    conanfile = GenConanfile().with_settings("os", "arch")\
        .with_generator("CMakeToolchain")
    profile = textwrap.dedent("""
    [settings]
    os=Android
    os.api_level=23
    arch=armv8

    [conf]
    tools.android:ndk_path=/foo
    tools.build:cflags=["-foobar"]
    tools.build:cxxflags=["-barfoo"]
    """)
    if cmake_legacy_toolchain is not None:
        profile += f"\ntools.android:cmake_legacy_toolchain={cmake_legacy_toolchain}"

    client.save({"conanfile.py": conanfile, "profile_host": profile})
    client.run("install . -pr profile_host")
    warning_text = "Consider setting tools.android:cmake_legacy_toolchain to False"
    if cmake_legacy_toolchain is not False:
        assert warning_text in client.out
    else:
        assert warning_text not in client.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Only Windows")
def test_presets_paths_normalization():
    # https://github.com/conan-io/conan/issues/11795
    # But then also https://github.com/conan-io/conan/issues/18434
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Conan(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            generators = "CMakeToolchain"

            def layout(self):
                cmake_layout(self)
        """)
    client.save({"conanfile.py": conanfile, "CMakeLists.txt": "foo"})
    client.run("install .")

    presets = json.loads(client.load("CMakeUserPresets.json"))
    assert "/" in presets["include"][0]
    assert "\\" not in presets["include"][0]


@pytest.mark.parametrize("arch, arch_toolset", [("x86", "x86_64"), ("x86_64", "x86_64"),
                                                ("x86", "x86"), ("x86_64", "x86")])
def test_presets_ninja_msvc(arch, arch_toolset):
    client = TestClient()
    conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.cmake import cmake_layout

            class Conan(ConanFile):
                settings = "os", "arch", "compiler", "build_type"
                generators = "CMakeToolchain"

                def layout(self):
                    cmake_layout(self)
            """)
    client.save({"conanfile.py": conanfile, "CMakeLists.txt": "foo"})
    configs = ["-c tools.cmake.cmaketoolchain:toolset_arch={}".format(arch_toolset),
               "-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.compiler.cppstd\"]'",
               "-c tools.cmake.cmaketoolchain:generator=Ninja"]
    msvc = " -s compiler=msvc -s compiler.version=191 -s compiler.runtime=static " \
           "-s compiler.runtime_type=Release"
    client.run("install . {} -s compiler.cppstd=14 {} -s arch={}".format(" ".join(configs), msvc, arch))

    presets = json.loads(client.load("build/14/Release/generators/CMakePresets.json"))

    toolset_value = {"x86_64": "v141,host=x86_64", "x86": "v141,host=x86"}.get(arch_toolset)
    arch_value = {"x86_64": "x64", "x86": "x86"}.get(arch)
    assert presets["configurePresets"][0]["architecture"]["value"] == arch_value
    assert presets["configurePresets"][0]["architecture"]["strategy"] == "external"
    assert presets["configurePresets"][0]["toolset"]["value"] == toolset_value
    assert presets["configurePresets"][0]["toolset"]["strategy"] == "external"

    # Only for Ninja, no ninja, no values
    rmdir(os.path.join(client.current_folder, "build"))
    configs = ["-c tools.cmake.cmaketoolchain:toolset_arch={}".format(arch_toolset),
               "-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.compiler.cppstd\"]'"]
    client.run(
        "install . {} -s compiler.cppstd=14 {} -s arch={}".format(" ".join(configs), msvc, arch))

    toolset_value = {"x86_64": "v141,host=x86_64", "x86": "v141,host=x86"}.get(arch_toolset)
    arch_value = {"x86_64": "x64", "x86": "Win32"}.get(arch)  # NOTE: Win32 is different!!
    presets = json.loads(client.load("build/14/generators/CMakePresets.json"))
    assert presets["configurePresets"][0]["architecture"]["value"] == arch_value
    assert presets["configurePresets"][0]["architecture"]["strategy"] == "external"
    assert presets["configurePresets"][0]["toolset"]["value"] == toolset_value
    assert presets["configurePresets"][0]["toolset"]["strategy"] == "external"

    rmdir(os.path.join(client.current_folder, "build"))
    configs = ["-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.compiler.cppstd\"]'",
               "-c tools.cmake.cmaketoolchain:generator=Ninja"]

    client.run(
        "install . {} -s compiler.cppstd=14 {} -s arch={}".format(" ".join(configs), msvc, arch))
    presets = json.loads(client.load("build/14/Release/generators/CMakePresets.json"))
    toolset_value = {"x86_64": "v141", "x86": "v141"}.get(arch_toolset)
    arch_value = {"x86_64": "x64", "x86": "x86"}.get(arch)
    assert presets["configurePresets"][0]["architecture"]["value"] == arch_value
    assert presets["configurePresets"][0]["architecture"]["strategy"] == "external"
    assert presets["configurePresets"][0]["toolset"]["value"] == toolset_value
    assert presets["configurePresets"][0]["toolset"]["strategy"] == "external"


def test_pkg_config_block():
    os_ = platform.system()
    os_ = "Macos" if os_ == "Darwin" else os_
    profile = textwrap.dedent("""
        [settings]
        os=%s
        arch=x86_64

        [conf]
        tools.gnu:pkg_config=/usr/local/bin/pkg-config
        """ % os_)

    client = TestClient(path_with_spaces=False)
    conanfile = GenConanfile().with_settings("os", "arch")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                "profile": profile})
    client.run("install . -pr:b profile -pr:h profile")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'set(PKG_CONFIG_EXECUTABLE /usr/local/bin/pkg-config CACHE FILEPATH ' in toolchain
    pathsep = ":" if os_ != "Windows" else ";"
    pkg_config_path_set = 'set(ENV{PKG_CONFIG_PATH} "%s$ENV{PKG_CONFIG_PATH}")' % \
                          ("${CMAKE_CURRENT_LIST_DIR}" + pathsep)
    assert pkg_config_path_set in toolchain


@pytest.mark.parametrize("path", ["subproject", False])
def test_user_presets_custom_location(path):
    client = TestClient()
    conanfile = textwrap.dedent("""
                import os
                from conan import ConanFile
                from conan.tools.cmake import cmake_layout, CMakeToolchain

                class Conan(ConanFile):
                    settings = "os", "arch", "compiler", "build_type"

                    def generate(self):
                        t = CMakeToolchain(self)
                        t.user_presets_path = {}
                        t.generate()

                    def layout(self):
                        cmake_layout(self)
                """.format('"{}"'.format(path) if isinstance(path, str) else path))
    client.save({"CMakeLists.txt": "",
                 "subproject/CMakeLists.txt": "",
                 "subproject2/foo.txt": "",
                 "conanfile.py": conanfile})

    # We want to generate it to build the subproject
    client.run("install . ")

    if path is not False:
        assert not os.path.exists(os.path.join(client.current_folder, "CMakeUserPresets.json"))
        assert os.path.exists(os.path.join(client.current_folder, "subproject", "CMakeUserPresets.json"))
    else:
        assert not os.path.exists(os.path.join(client.current_folder, "CMakeUserPresets.json"))
        assert not os.path.exists(os.path.join(client.current_folder, "False", "CMakeUserPresets.json"))


def test_set_cmake_lang_compilers_and_launchers():
    profile = textwrap.dedent(r"""
    [settings]
    os=Windows
    arch=x86_64
    compiler=clang
    compiler.version=15
    compiler.libcxx=libstdc++11
    [conf]
    tools.build:compiler_executables={"c": "/my/local/gcc", "cpp": "g++", "rc": "C:\\local\\rc.exe"}
    """)
    client = TestClient(path_with_spaces=False)
    conanfile = GenConanfile().with_settings("os", "arch", "compiler")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                "profile": profile})
    client.run("install . -pr:b profile -pr:h profile")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'set(CMAKE_C_COMPILER clang)' not in toolchain
    assert 'set(CMAKE_CXX_COMPILER clang++)' not in toolchain
    assert 'set(CMAKE_C_COMPILER "/my/local/gcc")' in toolchain
    assert 'set(CMAKE_CXX_COMPILER "g++")' in toolchain
    assert 'set(CMAKE_RC_COMPILER "C:/local/rc.exe")' in toolchain


def test_cmake_layout_toolchain_folder():
    """ in single-config generators, the toolchain is a different file per configuration
    https://github.com/conan-io/conan/issues/12827
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Conan(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            generators = "CMakeToolchain"

            def layout(self):
                cmake_layout(self)
            """)
    c.save({"conanfile.py": conanfile})
    c.run("install . -s os=Linux -s compiler=gcc -s compiler.version=7 -s build_type=Release "
          "-s compiler.libcxx=libstdc++11")
    assert os.path.exists(os.path.join(c.current_folder,
                                       "build/Release/generators/conan_toolchain.cmake"))
    c.run("install . -s os=Linux -s compiler=gcc -s compiler.version=7 -s build_type=Debug "
          "-s compiler.libcxx=libstdc++11")
    assert os.path.exists(os.path.join(c.current_folder,
                                       "build/Debug/generators/conan_toolchain.cmake"))
    c.run("install . -s os=Linux -s compiler=gcc -s compiler.version=7 -s build_type=Debug "
          "-s compiler.libcxx=libstdc++11 -s arch=x86 "
          "-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.arch\", \"settings.build_type\"]'")
    assert os.path.exists(os.path.join(c.current_folder,
                                       "build/x86-debug/generators/conan_toolchain.cmake"))
    c.run("install . -s os=Linux -s compiler=gcc -s compiler.version=7 -s build_type=Debug "
          "-s compiler.libcxx=libstdc++11 -s arch=x86 "
          "-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.os\"]'")
    assert os.path.exists(os.path.join(c.current_folder,
                                       "build/linux/Debug/generators/conan_toolchain.cmake"))


def test_build_folder_vars_editables():
    """ when packages are in editable, they must also follow the build_folder_vars
    https://github.com/conan-io/conan/issues/13485
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Conan(ConanFile):
            name = "dep"
            version = "0.1"
            settings = "os", "build_type"
            generators = "CMakeToolchain"

            def layout(self):
                cmake_layout(self)
        """)
    c.save({"dep/conanfile.py": conanfile,
            "app/conanfile.py": GenConanfile().with_requires("dep/0.1")})
    c.run("editable add dep")
    conf = "tools.cmake.cmake_layout:build_folder_vars='[\"settings.os\", \"settings.build_type\"]'"
    settings = " -s os=FreeBSD -s arch=armv8 -s build_type=Debug"
    c.run("install app -c={} {} --build=editable".format(conf, settings))
    assert os.path.exists(os.path.join(c.current_folder, "dep", "build", "freebsd-debug"))


def test_set_linker_scripts():
    profile = textwrap.dedent(r"""
    [settings]
    os=Windows
    arch=x86_64
    compiler=clang
    compiler.version=15
    compiler.libcxx=libstdc++11
    [conf]
    tools.build:linker_scripts=["/usr/local/src/flash.ld", "C:\\local\\extra_data.ld"]
    """)
    client = TestClient(path_with_spaces=False)
    conanfile = GenConanfile().with_settings("os", "arch", "compiler")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                "profile": profile})
    client.run("install . -pr:b profile -pr:h profile")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'string(APPEND CONAN_EXE_LINKER_FLAGS ' \
           r'" -T\"/usr/local/src/flash.ld\" -T\"C:/local/extra_data.ld\"")' in toolchain


def test_test_package_layout():
    """
    test that the ``test_package`` folder also follows the cmake_layout and the
    build_folder_vars
    """
    client = TestClient()
    test_conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Conan(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            generators = "CMakeToolchain"

            def requirements(self):
                self.requires(self.tested_reference_str)

            def layout(self):
                cmake_layout(self)

            def test(self):
                pass
    """)
    client.save({"conanfile.py": GenConanfile("pkg", "0.1"),
                 "test_package/conanfile.py": test_conanfile})
    config = "-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.compiler.cppstd\"]'"
    client.run(f"create . {config} -s compiler.cppstd=14")
    build_folder = client.created_test_build_folder("pkg/0.1")
    assert os.path.exists(os.path.join(client.current_folder, "test_package", build_folder))
    client.run(f"create . {config} -s compiler.cppstd=17")
    build_folder2 = client.created_test_build_folder("pkg/0.1")
    assert os.path.exists(os.path.join(client.current_folder, "test_package", build_folder2))
    assert build_folder != build_folder2


def test_presets_not_found_error_msg():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class Conan(ConanFile):
            settings = "build_type"

            def build(self):
                CMake(self).configure()
    """)
    client.save({"conanfile.py": conanfile})
    client.run("build .", assert_error=True)
    assert "CMakePresets.json was not found" in client.out
    assert "Check that you are using CMakeToolchain as generator " \
           "to ensure its correct initialization." in client.out


def test_recipe_build_folders_vars():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Conan(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "arch", "build_type"
            options = {"shared": [True, False]}
            generators = "CMakeToolchain"

            def layout(self):
                self.folders.build_folder_vars = ["settings.os", "options.shared"]
                cmake_layout(self)
        """)
    client.save({"conanfile.py": conanfile})
    client.run("install . -s os=Windows -s arch=armv8 -s build_type=Debug -o shared=True")
    presets = client.load("build/windows-shared/Debug/generators/CMakePresets.json")
    assert "conan-windows-shared-debug" in presets
    client.run("install . -s os=Linux -s arch=x86 -s build_type=Release -o shared=False")
    presets = client.load("build/linux-static/Release/generators/CMakePresets.json")
    assert "linux-static-release" in presets

    # CLI override has priority
    client.run("install . -s os=Linux -s arch=x86 -s build_type=Release -o shared=False "
               "-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.os\"]'")
    presets = client.load("build/linux/Release/generators/CMakePresets.json")
    assert "conan-linux-release" in presets

    # Now we do the build in the cache, the recipe folders are still used
    client.run("create . -s os=Windows -s arch=armv8 -s build_type=Debug -o shared=True")
    build_folder = client.created_layout().build()
    presets = load(os.path.join(build_folder,
                                "build/windows-shared/Debug/generators/CMakePresets.json"))
    assert "conan-windows-shared-debug" in presets

    # If we change the conf ``build_folder_vars``, it doesn't affect the cache build
    client.run("create . -s os=Windows -s arch=armv8 -s build_type=Debug -o shared=True "
               "-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.os\"]'")
    build_folder = client.created_layout().build()
    presets = load(os.path.join(build_folder,
                                "build/windows-shared/Debug/generators/CMakePresets.json"))
    assert "conan-windows-shared-debug" in presets


def test_build_folder_vars_self_name_version():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Conan(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "build_type"
            generators = "CMakeToolchain"

            def layout(self):
                self.folders.build_folder_vars = ["settings.os", "self.name", "self.version"]
                cmake_layout(self)
        """)
    client.save({"conanfile.py": conanfile})
    client.run("install . -s os=Windows -s build_type=Debug")
    presets = client.load("build/windows-pkg-0.1/Debug/generators/CMakePresets.json")
    assert "conan-windows-pkg-0.1-debug" in presets
    client.run("install . -s os=Linux -s build_type=Release")
    presets = client.load("build/linux-pkg-0.1/Release/generators/CMakePresets.json")
    assert "linux-pkg-0.1-release" in presets

    # CLI override has priority
    client.run("install . -s os=Linux  -s build_type=Release "
               "-c tools.cmake.cmake_layout:build_folder_vars='[\"self.name\"]'")
    presets = client.load("build/pkg/Release/generators/CMakePresets.json")
    assert "conan-pkg-release" in presets

    # Now we do the build in the cache, the recipe folders are still used
    client.run("create . -s os=Windows -s build_type=Debug")
    build_folder = client.created_layout().build()
    presets = load(os.path.join(build_folder,
                                "build/windows-pkg-0.1/Debug/generators/CMakePresets.json"))
    assert "conan-windows-pkg-0.1-debug" in presets

    # If we change the conf ``build_folder_vars``, it doesn't affect the cache build
    client.run("create . -s os=Windows -s build_type=Debug "
               "-c tools.cmake.cmake_layout:build_folder_vars='[\"settings.os\"]'")
    build_folder = client.created_layout().build()
    presets = load(os.path.join(build_folder,
                                "build/windows-pkg-0.1/Debug/generators/CMakePresets.json"))
    assert "conan-windows-pkg-0.1-debug" in presets


def test_build_folder_vars_constants_user():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout

        class Conan(ConanFile):
            name = "dep"
            version = "0.1"
            settings = "os", "build_type"
            generators = "CMakeToolchain"

            def layout(self):
                cmake_layout(self)
        """)
    c.save({"conanfile.py": conanfile})
    conf = "tools.cmake.cmake_layout:build_folder_vars='[\"const.myvalue\"]'"
    settings = " -s os=FreeBSD -s arch=armv8 -s build_type=Debug"
    c.run("install . -c={} {}".format(conf, settings))
    assert "cmake --preset conan-myvalue-debug" in c.out
    assert os.path.exists(os.path.join(c.current_folder, "build", "myvalue", "Debug"))
    presets = load(os.path.join(c.current_folder,
                                "build/myvalue/Debug/generators/CMakePresets.json"))
    assert "conan-myvalue-debug" in presets


def test_extra_flags():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain

        class Conan(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "arch", "build_type"
            def generate(self):
                tc = CMakeToolchain(self)
                tc.extra_cxxflags = ["extra_cxxflags"]
                tc.extra_cflags = ["extra_cflags"]
                tc.extra_sharedlinkflags = ["extra_sharedlinkflags"]
                tc.extra_exelinkflags = ["extra_exelinkflags"]
                tc.generate()
        """)
    profile = textwrap.dedent("""
        include(default)
        [conf]
        tools.build:cxxflags+=['cxxflags']
        tools.build:cflags+=['cflags']
        tools.build:sharedlinkflags+=['sharedlinkflags']
        tools.build:exelinkflags+=['exelinkflags']
        """)
    client.save({"conanfile.py": conanfile, "profile": profile})
    client.run('install . -pr=./profile')
    toolchain = client.load("conan_toolchain.cmake")

    assert 'string(APPEND CONAN_CXX_FLAGS " extra_cxxflags cxxflags")' in toolchain
    assert 'string(APPEND CONAN_C_FLAGS " extra_cflags cflags")' in toolchain
    assert 'string(APPEND CONAN_SHARED_LINKER_FLAGS " extra_sharedlinkflags sharedlinkflags")' in toolchain
    assert 'string(APPEND CONAN_EXE_LINKER_FLAGS " extra_exelinkflags exelinkflags")' in toolchain


def test_avoid_ovewrite_user_cmakepresets():
    # https://github.com/conan-io/conan/issues/15052
    c = TestClient()
    c.save({"conanfile.txt": "",
            "CMakePresets.json": "{}"})
    c.run('install . -g CMakeToolchain', assert_error=True)
    assert "Error in generator 'CMakeToolchain': Existing CMakePresets.json not generated" in c.out
    assert "Use --output-folder or define a 'layout' to avoid collision" in c.out


def test_presets_njobs():
    c = TestClient()
    c.save({"conanfile.txt": ""})
    c.run('install . -g CMakeToolchain -c tools.build:jobs=42')
    presets = json.loads(c.load("CMakePresets.json"))
    assert presets["buildPresets"][0]["jobs"] == 42
    assert presets["testPresets"][0]["execution"]["jobs"] == 42
    c.run('install . -g CMakeToolchain -c tools.build:jobs=0')
    presets = json.loads(c.load("CMakePresets.json"))
    assert "jobs" not in presets["buildPresets"][0]
    assert "execution" not in presets["testPresets"][0]


def test_add_cmakeexe_to_presets():
    c = TestClient()

    tool = textwrap.dedent(r"""
        import os
        from conan import ConanFile
        from conan.tools.files import chdir, save
        class Tool(ConanFile):
            name = "cmake"
            version = "3.27"
            settings = "os", "compiler", "arch", "build_type"
            def package(self):
                with chdir(self, self.package_folder):
                    save(self, "bin/{}", "")
        """)

    profile = textwrap.dedent("""
        include(default)
        [platform_tool_requires]
        cmake/3.27
    """)

    consumer = textwrap.dedent("""
        [tool_requires]
        cmake/3.27
        [layout]
        cmake_layout
    """)

    cmake_exe = "cmake.exe" if platform.system() == "Windows" else "cmake"

    c.save({"tool.py": tool.format(cmake_exe),
            "conanfile.txt": consumer,
            "myprofile": profile})
    c.run("create tool.py")
    c.run("install . -g CMakeToolchain -g CMakeDeps")

    presets_path = os.path.join("build", "Release", "generators", "CMakePresets.json") \
        if platform.system() != "Windows" else os.path.join("build", "generators", "CMakePresets.json")
    presets = json.loads(c.load(presets_path))

    assert cmake_exe == os.path.basename(presets["configurePresets"][0].get("cmakeExecutable"))

    # if we set "tools.cmake:cmake_program" that will have preference
    c.run("install . -g CMakeToolchain -g CMakeDeps -c tools.cmake:cmake_program='/other/path/cmake'")
    presets = json.loads(c.load(presets_path))

    assert '/other/path/cmake' == presets["configurePresets"][0].get("cmakeExecutable")

    # if we have a platform_tool_requires it will not be set because  it is filtered before
    # so it will not be in direct_build dependencies
    c.run("install . -g CMakeToolchain -g CMakeDeps -pr:h=./myprofile")

    presets = json.loads(c.load(presets_path))
    assert presets["configurePresets"][0].get("cmakeExecutable") is None


def test_toolchain_ends_newline():
    # https://github.com/conan-io/conan/issues/15785
    client = TestClient()
    client.save({"conanfile.py": GenConanfile()})
    client.run("install . -g CMakeToolchain")
    toolchain = client.load("conan_toolchain.cmake")
    assert toolchain[-1] == "\n"


def test_toolchain_and_compilers_build_context():
    """
    Tests how CMakeToolchain manages the build context profile if the build profile is
    specifying another compiler path (using conf)

    Issue related: https://github.com/conan-io/conan/issues/15878
    """
    host = textwrap.dedent("""
    [settings]
    arch=armv8
    build_type=Release
    compiler=gcc
    compiler.cppstd=gnu17
    compiler.libcxx=libstdc++11
    compiler.version=11
    os=Linux

    [conf]
    tools.build:compiler_executables={"c": "gcc", "cpp": "g++"}
    """)
    build = textwrap.dedent("""
    [settings]
    os=Linux
    arch=x86_64
    compiler=clang
    compiler.version=12
    compiler.libcxx=libc++
    compiler.cppstd=11

    [conf]
    tools.build:compiler_executables={"c": "clang", "cpp": "clang++"}
    """)
    tool = textwrap.dedent("""
    import os
    from conan import ConanFile
    from conan.tools.files import load

    class toolRecipe(ConanFile):
        name = "tool"
        version = "1.0"
        # Binary configuration
        settings = "os", "compiler", "build_type", "arch"
        generators = "CMakeToolchain"

        def build(self):
            toolchain = os.path.join(self.generators_folder, "conan_toolchain.cmake")
            content = load(self, toolchain)
            assert 'set(CMAKE_C_COMPILER "clang")' in content
            assert 'set(CMAKE_CXX_COMPILER "clang++")' in content
    """)
    consumer = textwrap.dedent("""
    import os
    from conan import ConanFile
    from conan.tools.files import load

    class consumerRecipe(ConanFile):
        name = "consumer"
        version = "1.0"
        # Binary configuration
        settings = "os", "compiler", "build_type", "arch"
        generators = "CMakeToolchain"
        tool_requires = "tool/1.0"

        def build(self):
            toolchain = os.path.join(self.generators_folder, "conan_toolchain.cmake")
            content = load(self, toolchain)
            assert 'set(CMAKE_C_COMPILER "gcc")' in content
            assert 'set(CMAKE_CXX_COMPILER "g++")' in content
    """)
    client = TestClient()
    client.save({
        "host": host,
        "build": build,
        "tool/conanfile.py": tool,
        "consumer/conanfile.py": consumer
    })
    client.run("export tool")
    client.run("create consumer -pr:h host -pr:b build --build=missing")


def test_toolchain_keep_absolute_paths():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain, cmake_layout
        class Pkg(ConanFile):
            settings = "build_type"
            def generate(self):
                tc = CMakeToolchain(self)
                tc.absolute_paths = True
                tc.generate()
            def layout(self):
                cmake_layout(self)
        """)
    c.save({"conanfile.py": conanfile,
            "CMakeLists.txt": ""})
    c.run('install . ')

    user_presets = json.loads(c.load("CMakeUserPresets.json"))
    assert os.path.isabs(user_presets["include"][0])
    presets = json.loads(c.load(user_presets["include"][0]))
    assert os.path.isabs(presets["configurePresets"][0]["toolchainFile"])


def test_customize_cmakeuserpresets():
    # https://github.com/conan-io/conan/issues/15639
    c = TestClient()
    c.save({"conanfile.py": GenConanfile(),
            "CMakeLists.txt": ""})
    c.run("install . -g CMakeToolchain -of=build")
    old = c.load("CMakeUserPresets.json")
    os.remove(os.path.join(c.current_folder, "CMakeUserPresets.json"))

    c.run("install . -g CMakeToolchain -of=build -c tools.cmake.cmaketoolchain:user_presets=my.json")
    new = c.load("my.json")
    assert old == new
    # also in subfolder
    c.run("install . -g CMakeToolchain -of=build "
          "-c tools.cmake.cmaketoolchain:user_presets=mysub/my.json")
    assert os.path.exists(os.path.join(c.current_folder, "mysub", "my.json"))

    assert not os.path.exists(os.path.join(c.current_folder, "CMakeUserPresets.json"))
    c.run("install . -g CMakeToolchain -of=build -c tools.cmake.cmaketoolchain:user_presets=")
    assert not os.path.exists(os.path.join(c.current_folder, "CMakeUserPresets.json"))


def test_output_dirs_gnudirs_local_default():
    # https://github.com/conan-io/conan/issues/14733
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout
        from conan.tools.files import load

        class Conan(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "arch", "compiler", "build_type"
            generators = "CMakeToolchain"
            def build(self):
                tc = load(self, "conan_toolchain.cmake")
                self.output.info(tc)
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .")

    def _assert_install(out):
        assert 'set(CMAKE_INSTALL_BINDIR "bin")' in out
        assert 'set(CMAKE_INSTALL_SBINDIR "bin")' in out
        assert 'set(CMAKE_INSTALL_LIBEXECDIR "bin")' in out
        assert 'set(CMAKE_INSTALL_LIBDIR "lib")' in out
        assert 'set(CMAKE_INSTALL_INCLUDEDIR "include")' in out

    _assert_install(c.out)
    assert "CMAKE_INSTALL_PREFIX" in c.out

    c.run("build .")
    _assert_install(c.out)
    assert "CMAKE_INSTALL_PREFIX" not in c.out


def test_output_dirs_gnudirs_local_custom():
    # https://github.com/conan-io/conan/issues/14733
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import cmake_layout
        from conan.tools.files import load

        class Conan(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "arch", "compiler", "build_type"
            generators = "CMakeToolchain"
            def layout(self):
                self.cpp.package.bindirs = ["mybindir"]
                self.cpp.package.includedirs = ["myincludedir"]
                self.cpp.package.libdirs = ["mylibdir"]

            def build(self):
                tc = load(self, "conan_toolchain.cmake")
                self.output.info(tc)
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .")

    def _assert_install(out):
        assert 'set(CMAKE_INSTALL_BINDIR "mybindir")' in out
        assert 'set(CMAKE_INSTALL_SBINDIR "mybindir")' in out
        assert 'set(CMAKE_INSTALL_LIBEXECDIR "mybindir")' in out
        assert 'set(CMAKE_INSTALL_LIBDIR "mylibdir")' in out
        assert 'set(CMAKE_INSTALL_INCLUDEDIR "myincludedir")' in out

    _assert_install(c.out)
    assert "CMAKE_INSTALL_PREFIX" in c.out

    c.run("build .")
    _assert_install(c.out)
    assert "CMAKE_INSTALL_PREFIX" not in c.out


def test_toolchain_extra_variables():
    windows_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        [conf]
        tools.cmake.cmaketoolchain:extra_variables={'CMAKE_GENERATOR_INSTANCE': '${GENERATOR_INSTANCE}/buildTools/', 'FOO': '42' }
        """)

    client = TestClient()
    client.save({"conanfile.txt": "[generators]\nCMakeToolchain",
                 "windows": windows_profile})

    # Test passing extra_variables from pro ile
    client.run("install . --profile:host=windows")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'set(CMAKE_GENERATOR_INSTANCE "${GENERATOR_INSTANCE}/buildTools/")' in toolchain
    assert 'set(FOO "42")' in toolchain

    # Test input from command line passing dict between doble quotes
    client.run(textwrap.dedent(r"""
        install . -c tools.cmake.cmaketoolchain:extra_variables="{'CMAKE_GENERATOR_INSTANCE': '${GENERATOR_INSTANCE}/buildTools/', 'FOO': 42.2, 'DICT': {'value': 1}, 'CACHE_VAR': {'value': 'hello world', 'cache': True, 'type': 'BOOL', 'docstring': 'test variable'}}"
        """)
    )

    toolchain = client.load("conan_toolchain.cmake")
    assert 'set(CMAKE_GENERATOR_INSTANCE "${GENERATOR_INSTANCE}/buildTools/")' in toolchain
    assert 'set(FOO 42.2)' in toolchain
    assert 'set(DICT 1)' in toolchain
    assert 'set(CACHE_VAR "hello world" CACHE BOOL "test variable")' in toolchain

    client.run(textwrap.dedent("""
        install . -c tools.cmake.cmaketoolchain:extra_variables="{'myVar': {'value': 'hello world', 'cache': 'true'}}"
    """), assert_error=True)
    assert 'tools.cmake.cmaketoolchain:extra_variables "myVar" "cache" must be a boolean' in client.out

    # Test invalid force
    client.run(textwrap.dedent("""
        install . -c tools.cmake.cmaketoolchain:extra_variables="{'myVar': {'value': 'hello world', 'force': True}}"
    """), assert_error=True)
    assert 'tools.cmake.cmaketoolchain:extra_variables "myVar" "force" is only allowed for cache variables' in client.out

    client.run(textwrap.dedent("""
        install . -c tools.cmake.cmaketoolchain:extra_variables="{'myVar': {'value': 'hello world', 'cache': True, 'force': 'true'}}"
    """), assert_error=True)
    assert 'tools.cmake.cmaketoolchain:extra_variables "myVar" "force" must be a boolean' in client.out

    # Test invalid cache variable
    client.run(textwrap.dedent("""
        install . -c tools.cmake.cmaketoolchain:extra_variables="{'myVar': {'value': 'hello world', 'cache': True}}"
    """), assert_error=True)
    assert 'tools.cmake.cmaketoolchain:extra_variables "myVar" needs "type" defined for cache variable' in client.out

    client.run(textwrap.dedent("""
        install . -c tools.cmake.cmaketoolchain:extra_variables="{'myVar': {'value': 'hello world', 'cache': True, 'type': 'INVALID_TYPE'}}"
    """), assert_error=True)
    assert 'tools.cmake.cmaketoolchain:extra_variables "myVar" invalid type "INVALID_TYPE" for cache variable. Possible types: BOOL, FILEPATH, PATH, STRING, INTERNAL' in client.out

    client.run(textwrap.dedent("""
        install . -c tools.cmake.cmaketoolchain:extra_variables="{'CACHE_VAR_DEFAULT_DOC': {'value': 'hello world', 'cache': True, 'type': 'PATH'}}"
    """))
    toolchain = client.load("conan_toolchain.cmake")
    assert 'set(CACHE_VAR_DEFAULT_DOC "hello world" CACHE PATH "CACHE_VAR_DEFAULT_DOC")' in toolchain

    client.run(textwrap.dedent("""
        install . -c tools.cmake.cmaketoolchain:extra_variables="{'myVar': {'value': 'hello world', 'cache': True, 'type': 'PATH', 'docstring': 'My cache variable', 'force': True}}"
    """))
    toolchain = client.load("conan_toolchain.cmake")
    assert 'set(myVar "hello world" CACHE PATH "My cache variable" FORCE)' in toolchain


def test_variables_wrong_scaping():
    # https://github.com/conan-io/conan/issues/16432
    c = TestClient()
    c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
            "pkg/conanfile.txt": "[tool_requires]\ntool/0.1\n[generators]\nCMakeToolchain"})
    c.run("create tool")
    c.run("install pkg")
    toolchain = c.load("pkg/conan_toolchain.cmake")
    cache_folder = c.cache_folder.replace("\\", "/")
    assert f'list(PREPEND CMAKE_PROGRAM_PATH "{cache_folder}' in toolchain

    c.run("install pkg --deployer=full_deploy")
    toolchain = c.load("pkg/conan_toolchain.cmake")
    assert 'list(PREPEND CMAKE_PROGRAM_PATH "${CMAKE_CURRENT_LIST_DIR}/full_deploy' in toolchain


def test_tricore():
    # making sure the arch ``tc131`` is there
    c = TestClient()
    c.save({"conanfile.txt": "[generators]\nCMakeToolchain"})
    c.run("install . -s os=baremetal -s compiler=gcc -s arch=tc131")
    content = c.load("conan_toolchain.cmake")
    assert 'set(CMAKE_SYSTEM_NAME Generic-ELF)' in content
    assert 'set(CMAKE_SYSTEM_PROCESSOR tricore)' in content
    assert 'string(APPEND CONAN_CXX_FLAGS " -mtc131")' in content
    assert 'string(APPEND CONAN_C_FLAGS " -mtc131")' in content
    assert 'string(APPEND CONAN_SHARED_LINKER_FLAGS " -mtc131")' in content
    assert 'string(APPEND CONAN_EXE_LINKER_FLAGS " -mtc131")' in content


def test_declared_stdlib_and_passed():
    client = TestClient()
    client.save({"conanfile.txt": "[generators]\nCMakeToolchain"})

    client.run('install . -s compiler=sun-cc -s compiler.libcxx=libCstd')
    tc = client.load("conan_toolchain.cmake")
    assert 'string(APPEND CONAN_CXX_FLAGS " -library=Cstd")' in tc

    client.run('install . -s compiler=sun-cc -s compiler.libcxx=libstdcxx')
    tc = client.load("conan_toolchain.cmake")
    assert 'string(APPEND CONAN_CXX_FLAGS " -library=stdcxx4")' in tc


def test_cmake_presets_compiler():
    profile = textwrap.dedent(r"""
        [settings]
        os=Windows
        arch=x86_64
        compiler=msvc
        compiler.version=193
        compiler.runtime=dynamic
        [conf]
        tools.build:compiler_executables={"c": "cl", "cpp": "cl.exe"}
        """)
    client = TestClient()
    conanfile = GenConanfile().with_settings("os", "arch", "compiler")\
        .with_generator("CMakeToolchain")
    client.save({"conanfile.py": conanfile,
                 "profile": profile})
    client.run("install . -pr:b profile -pr:h profile")
    presets = json.loads(client.load("CMakePresets.json"))
    cache_variables = presets["configurePresets"][0]["cacheVariables"]
    # https://github.com/microsoft/vscode-cmake-tools/blob/a1ceda25ea93fc0060324de15970a8baa69addf6/src/presets/preset.ts#L1095C23-L1095C35
    assert cache_variables["CMAKE_C_COMPILER"] == "cl"
    assert cache_variables["CMAKE_CXX_COMPILER"] == "cl.exe"


@pytest.mark.parametrize(
    "threads, flags",
    [("posix", "-pthread"), ("wasm_workers", "-sWASM_WORKERS=1")],
)
def test_thread_flags(threads, flags):
    client = TestClient()
    profile = textwrap.dedent(f"""
        [settings]
        arch=wasm
        build_type=Release
        compiler=emcc
        compiler.cppstd=17
        compiler.threads={threads}
        compiler.libcxx=libc++
        compiler.version=4.0.10
        os=Emscripten
        """)
    client.save(
        {
            "conanfile.py": GenConanfile("pkg", "1.0")
            .with_settings("os", "arch", "compiler", "build_type")
            .with_generator("CMakeToolchain"),
            "profile": profile,
        }
    )
    client.run("install . -pr=./profile")
    toolchain = client.load("conan_toolchain.cmake")
    assert f'string(APPEND CONAN_CXX_FLAGS " {flags}")' in toolchain
    assert f'string(APPEND CONAN_C_FLAGS " {flags}")' in toolchain
    assert f'string(APPEND CONAN_SHARED_LINKER_FLAGS " {flags}")' in toolchain
    assert f'string(APPEND CONAN_EXE_LINKER_FLAGS " {flags}")' in toolchain



from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_editable_ranges():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile("dep"),
            "dep2/conanfile.py": GenConanfile("dep", "0.2"),
            "other/conanfile.py": GenConanfile("other", "23.0"),  # shouldn't be resolved!
            "app/conanfile.py": GenConanfile("app", "0.1").with_requires("dep/[>=0.1]")})
    c.run("editable add other")  # This shouldn't be used
    c.run("editable add dep --version=0.1")  # this should be used
    c.run("editable add dep --version=2.0 --user=myuser --channel=mychannel")  # This shouldn't be used
    c.run("install app")
    c.assert_listed_require({"dep/0.1": "Editable"})
    assert "dep/[>=0.1]: dep/0.1" in c.out

    # new version, uses new one
    c.run("editable add dep2")
    c.run("install app")
    c.assert_listed_require({"dep/0.2": "Editable"})
    assert "dep/[>=0.1]: dep/0.2" in c.out
    assert "dep/0.1" not in c.out

    # If a newer one is in cache, it resolves to cache
    c.run("create dep --version=0.3")
    c.run("install app")
    c.assert_listed_require({"dep/0.3": "Cache"})
    assert "dep/[>=0.1]: dep/0.3" in c.out
    assert "dep/0.1" not in c.out



from collections import OrderedDict

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestServer


class TestSelectedRemotesInstall:
    @pytest.fixture(autouse=True)
    def _setup(self):
        servers = OrderedDict()
        for index in range(3):
            servers[f"server{index}"] = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")],
                                                   users={"user": "password"})
        self.client = TestClient(servers=servers, inputs=3 * ["user", "password"])

    def test_selected_remotes(self):
        self.client.save({"conanfile.py": GenConanfile("liba", "1.0").with_build_msg("OLDREV")})
        self.client.run("create .")
        self.client.run("upload liba/1.0 -r server0 -c")
        self.client.run("remove * -c")
        self.client.save({"conanfile.py": GenConanfile("liba", "1.0").with_build_msg("NEWER_REV")})
        self.client.run("create .")
        self.client.run("upload liba/1.0 -r server1 -c")
        self.client.run("remove * -c")

        self.client.run("install --requires=liba/1.0 -r server0 -r server1 --build='*'")
        # we install the revision from the server with more preference
        assert "OLDREV" in self.client.out
        self.client.run("remove * -c")
        self.client.run("install --requires=liba/1.0 -r server1 -r server0 --build='*'")
        # changing the order of the remotes in the arguments does change the result
        # we install the revision from the server with more preference
        assert "NEWER_REV" in self.client.out
        # select two remotes, just one has liba, will install the rev from that one
        self.client.run("remove * -c")
        self.client.run("install --requires=liba/1.0 -r server2 -r server1 --build='*'")
        assert "NEWER_REV" in self.client.out

        self.client.save({"consumer.py": GenConanfile().with_require("liba/1.0")})
        self.client.run("remove * -c")
        self.client.run("create . --build='*' -r server0 -r server1 -r server2")
        assert "NEWER_REV" in self.client.out

    def test_upload_raise_multiple_remotes(self):
        self.client.run("upload liba -r server0 -r server1 -c", assert_error=True)
        assert "conan upload: error: -r can only be specified once" in self.client.out

    def test_remove_raise_multiple_remotes(self):
        self.client.run("remove liba -r server0 -r server1 -c", assert_error=True)
        assert "conan remove: error: -r can only be specified once" in self.client.out



import json
import os
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_info_build_order():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": GenConanfile().with_requires("dep/0.1"),
            "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    c.run("export dep --name=dep --version=0.1")
    c.run("export pkg --name=pkg --version=0.1")
    c.run("graph build-order consumer --build=missing --format=json")
    bo_json = json.loads(c.stdout)

    result = [
        [
            {
                "ref": "dep/0.1#4d670581ccb765839f2239cc8dff8fbd",
                "depends": [],
                "packages": [[
                    {
                        "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                        'prev': None,
                        'filenames': [],
                        'info': {},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=dep/0.1 --build=dep/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            }
        ],
        [
            {
                "ref": "pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032",
                "depends": [
                    "dep/0.1#4d670581ccb765839f2239cc8dff8fbd"
                ],
                "packages": [[
                    {
                        "package_id": "59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                        'prev': None,
                        'filenames': [],
                        'info': {'requires': ['dep/0.1']},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=pkg/0.1 --build=pkg/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            }
        ]
    ]

    assert bo_json == result

    c.run("graph build-order consumer --order-by=recipe --build=missing --format=json")
    bo_json = json.loads(c.stdout)
    assert bo_json["order_by"] == "recipe"
    assert bo_json["order"] == result

    c.run("graph build-order consumer --build=missing --order-by=recipe --reduce --format=json")
    bo_json = json.loads(c.stdout)
    assert bo_json["order_by"] == "recipe"
    assert bo_json["order"] == result

    # test html format
    c.run("graph build-order consumer --build=missing --format=html")
    assert "<body>" in c.stdout
    c.run("graph build-order consumer --order-by=recipe --build=missing --format=html")
    assert "<body>" in c.stdout
    c.run("graph build-order consumer --order-by=configuration --build=missing --format=html")
    assert "<body>" in c.stdout


def test_info_build_order_configuration():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": GenConanfile().with_requires("dep/0.1"),
            "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    c.run("export dep --name=dep --version=0.1")
    c.run("export pkg --name=pkg --version=0.1")
    c.run("graph build-order consumer --build=missing --order=configuration --format=json")
    bo_json = json.loads(c.stdout)
    assert bo_json["order_by"] == "configuration"

    result = [
        [
            {
                "ref": "dep/0.1#4d670581ccb765839f2239cc8dff8fbd",
                "pref": "dep/0.1#4d670581ccb765839f2239cc8dff8fbd:da39a3ee5e6b4b0d3255bfef95601890afd80709",
                "depends": [],
                "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                'prev': None,
                'filenames': [],
                'info': {},
                "context": "host",
                "binary": "Build",
                'build_args': '--requires=dep/0.1 --build=dep/0.1',
                "options": [],
                "overrides": {}
            }
        ],
        [
            {
                "ref": "pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032",
                "pref": "pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032:59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                "depends": [
                    "dep/0.1#4d670581ccb765839f2239cc8dff8fbd:da39a3ee5e6b4b0d3255bfef95601890afd80709"
                ],
                "package_id": "59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                'prev': None,
                'filenames': [],
                'info': {'requires': ['dep/0.1']},
                "context": "host",
                "binary": "Build",
                'build_args': '--requires=pkg/0.1 --build=pkg/0.1',
                "options": [],
                "overrides": {}

            }
        ]
    ]

    assert bo_json["order"] == result

    c.run("graph build-order consumer --build=missing --order=configuration --reduce --format=json")
    bo_json = json.loads(c.stdout)
    assert bo_json["order"] == result


def test_info_build_order_configuration_text_formatter():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": GenConanfile().with_requires("dep/0.1"),
            "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    c.run("export dep --name=dep --version=0.1")
    c.run("export pkg --name=pkg --version=0.1")
    c.run("graph build-order consumer --build=missing --order=configuration --format=text")
    assert textwrap.dedent("""\
    ======== Computing the build order ========
    dep/0.1#4d670581ccb765839f2239cc8dff8fbd:da39a3ee5e6b4b0d3255bfef95601890afd80709 - Build
    pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032:59205ba5b14b8f4ebc216a6c51a89553021e82c1 - Build
    """) in c.out


def test_info_build_order_build_require():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": GenConanfile().with_tool_requires("dep/0.1"),
            "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    c.run("export dep --name=dep --version=0.1")
    c.run("export pkg --name=pkg --version=0.1")
    c.run("graph build-order  consumer --build=missing --format=json")
    bo_json = json.loads(c.stdout)
    result = [
        [
            {
                "ref": 'dep/0.1#4d670581ccb765839f2239cc8dff8fbd',
                "depends": [],
                "packages": [[
                    {
                        "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                        'prev': None,
                        'filenames': [],
                        'info': {},
                        "context": "build",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--tool-requires=dep/0.1 --build=dep/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            }
        ],
        [
            {
                "ref": "pkg/0.1#b5a40d7314ce57ebdcf8fa31257f3de1",
                "depends": [
                    "dep/0.1#4d670581ccb765839f2239cc8dff8fbd"
                ],
                "packages": [[
                    {
                        "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                        'prev': None,
                        'filenames': [],
                        'info': {},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=pkg/0.1 --build=pkg/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            }
        ]
    ]

    assert bo_json == result


def test_info_build_order_options():
    c = TestClient()
    # The normal default_options do NOT propagate to build_requires, it is necessary to use
    # self.requires(..., options=xxx)
    c.save({"tool/conanfile.py": GenConanfile().with_option("myopt", [1, 2, 3]),
            "dep1/conanfile.py": GenConanfile().with_tool_requirement("tool/0.1",
                                                                      options={"myopt": 1}),
            "dep2/conanfile.py": GenConanfile().with_tool_requirement("tool/0.1",
                                                                      options={"myopt": 2}),
            "consumer/conanfile.txt": "[requires]\ndep1/0.1\ndep2/0.1"})
    c.run("export tool --name=tool --version=0.1")
    c.run("export dep1 --name=dep1 --version=0.1")
    c.run("export dep2 --name=dep2 --version=0.1")

    c.run("graph build-order  consumer --build=missing --format=json")
    bo_json = json.loads(c.stdout)
    result = [
        [
            {'ref': 'tool/0.1#b4c19a1357b43877a2019dd2804336a9',
             'depends': [],
             'packages': [[
                 {'package_id': '1124b99dc8cd3c8bbf79121c7bf86ce40c725a40', 'prev': None,
                  'context': 'build', 'depends': [], "overrides": {},
                  'binary': 'Build', 'options': ['tool/0.1:myopt=2'], 'filenames': [],
                  'info': {'options': {'myopt': '2'}},
                  'build_args': '--tool-requires=tool/0.1 --build=tool/0.1 -o:b="tool/0.1:myopt=2"'},
                 {'package_id': 'a9035d84c5880b26c4b44acf70078c9a7dd37412', 'prev': None,
                  'context': 'build', 'depends': [], "overrides": {},
                  'info': {'options': {'myopt': '1'}},
                  'binary': 'Build', 'options': ['tool/0.1:myopt=1'],
                  'filenames': [],
                  'build_args': '--tool-requires=tool/0.1 --build=tool/0.1 -o:b="tool/0.1:myopt=1"'}
             ]]}
        ],
        [
            {'ref': 'dep1/0.1#7f0d80f9cb8c6bab06def7f6fb8f3b86',
             'depends': ['tool/0.1#b4c19a1357b43877a2019dd2804336a9'],
             'packages': [[
                 {'package_id': 'da39a3ee5e6b4b0d3255bfef95601890afd80709', 'prev': None,
                  'context': 'host', 'depends': [], 'binary': 'Build', 'options': [],
                  'filenames': [], 'info': {}, "overrides": {},
                  'build_args': '--requires=dep1/0.1 --build=dep1/0.1'}
             ]]},
            {'ref': 'dep2/0.1#23c789d2b36f0461e52cd6f139f97f5e',
             'depends': ['tool/0.1#b4c19a1357b43877a2019dd2804336a9'],
             'packages': [[
                 {'package_id': 'da39a3ee5e6b4b0d3255bfef95601890afd80709', 'prev': None,
                  'context': 'host', 'depends': [], 'binary': 'Build', 'options': [],
                  'filenames': [], 'info': {}, "overrides": {},
                  'build_args': '--requires=dep2/0.1 --build=dep2/0.1'}
             ]]}
        ]
    ]
    assert bo_json == result


def test_info_build_order_merge_multi_product():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": GenConanfile().with_requires("dep/0.1"),
            "consumer1/conanfile.txt": "[requires]\npkg/0.1",
            "consumer2/conanfile.txt": "[requires]\npkg/0.2"})
    c.run("export dep --name=dep --version=0.1")
    c.run("export pkg --name=pkg --version=0.1")
    c.run("export pkg --name=pkg --version=0.2")
    c.run("graph build-order consumer1  --build=missing --format=json", redirect_stdout="bo1.json")
    c.run("graph build-order consumer2  --build=missing --format=json", redirect_stdout="bo2.json")
    c.run("graph build-order-merge --file=bo1.json --file=bo2.json --format=json",
          redirect_stdout="bo3.json")

    bo_json = json.loads(c.load("bo3.json"))

    result = [
        [
            {
                "ref": "dep/0.1#4d670581ccb765839f2239cc8dff8fbd",
                "depends": [],
                "packages": [[
                    {
                        "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                        'prev': None,
                        'filenames': ["bo1", "bo2"],
                        'info': {},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=dep/0.1 --build=dep/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            }
        ],
        [
            {
                "ref": "pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032",
                "depends": [
                    "dep/0.1#4d670581ccb765839f2239cc8dff8fbd"
                ],
                "packages": [[
                    {
                        "package_id": "59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                        'prev': None,
                        'filenames': ["bo1"],
                        'info': {'requires': ['dep/0.1']},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=pkg/0.1 --build=pkg/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            },
            {
                "ref": "pkg/0.2#1ac8dd17c0f9f420935abd3b6a8fa032",
                "depends": [
                    "dep/0.1#4d670581ccb765839f2239cc8dff8fbd"
                ],
                "packages": [[
                    {
                        "package_id": "59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                        'prev': None,
                        'filenames': ["bo2"],
                        'info': {'requires': ['dep/0.1']},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=pkg/0.2 --build=pkg/0.2',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            }
        ]
    ]

    assert bo_json == result

    # test that html format for build-order-merge generates something
    c.run("graph build-order-merge --file=bo1.json --file=bo2.json --format=html")
    assert "<body>" in c.stdout


def test_info_build_order_merge_multi_product_configurations():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": GenConanfile().with_requires("dep/0.1"),
            "consumer1/conanfile.txt": "[requires]\npkg/0.1",
            "consumer2/conanfile.txt": "[requires]\npkg/0.2"})
    c.run("export dep --name=dep --version=0.1")
    c.run("export pkg --name=pkg --version=0.1")
    c.run("export pkg --name=pkg --version=0.2")
    c.run("graph build-order consumer1  --build=missing --order=configuration --format=json",
          redirect_stdout="bo1.json")
    c.run("graph build-order consumer2  --build=missing --order=configuration --format=json",
          redirect_stdout="bo2.json")
    c.run("graph build-order-merge --file=bo1.json --file=bo2.json --format=json",
          redirect_stdout="bo3.json")

    bo_json = json.loads(c.load("bo3.json"))
    assert bo_json["order_by"] == "configuration"
    result = [
        [
            {
                "ref": "dep/0.1#4d670581ccb765839f2239cc8dff8fbd",
                "pref": "dep/0.1#4d670581ccb765839f2239cc8dff8fbd:da39a3ee5e6b4b0d3255bfef95601890afd80709",
                "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                "prev": None,
                "context": "host",
                "binary": "Build",
                "options": [],
                "filenames": [
                    "bo1",
                    "bo2"
                ],
                'info': {},
                "depends": [],
                "overrides": {},
                "build_args": "--requires=dep/0.1 --build=dep/0.1"
            }
        ],
        [
            {
                "ref": "pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032",
                "pref": "pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032:59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                "package_id": "59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                "prev": None,
                "context": "host",
                "binary": "Build",
                "options": [],
                "filenames": [
                    "bo1"
                ],
                'info': {'requires': ['dep/0.1']},
                "depends": [
                    "dep/0.1#4d670581ccb765839f2239cc8dff8fbd:da39a3ee5e6b4b0d3255bfef95601890afd80709"
                ],
                "overrides": {},
                "build_args": "--requires=pkg/0.1 --build=pkg/0.1"
            },
            {
                "ref": "pkg/0.2#1ac8dd17c0f9f420935abd3b6a8fa032",
                "pref": "pkg/0.2#1ac8dd17c0f9f420935abd3b6a8fa032:59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                "package_id": "59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                "prev": None,
                "context": "host",
                "binary": "Build",
                "options": [],
                "filenames": [
                    "bo2"
                ],
                'info': {'requires': ['dep/0.1']},
                "depends": [
                    "dep/0.1#4d670581ccb765839f2239cc8dff8fbd:da39a3ee5e6b4b0d3255bfef95601890afd80709"
                ],
                "overrides": {},
                "build_args": "--requires=pkg/0.2 --build=pkg/0.2"
            }
        ]
    ]

    assert bo_json["order"] == result


def test_info_build_order_merge_conditionals():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def requirements(self):
                if self.settings.os == "Windows":
                    self.requires("depwin/[>0.0 <1.0]")
                else:
                    self.requires("depnix/[>0.0 <1.0]")
        """)
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": conanfile,
            "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    c.run("export dep --name=depwin --version=0.1")
    c.run("export dep --name=depnix --version=0.1")
    c.run("export pkg --name=pkg --version=0.1")
    c.run("graph build-order consumer --format=json --build=missing -s os=Windows",
          redirect_stdout="bo_win.json")
    c.run("graph build-order consumer --format=json --build=missing -s os=Linux",
          redirect_stdout="bo_nix.json")
    c.run("graph build-order-merge --file=bo_win.json --file=bo_nix.json --format=json",
          redirect_stdout="bo3.json")

    bo_json = json.loads(c.load("bo3.json"))

    result = [
        [
            {
                "ref": "depwin/0.1#4d670581ccb765839f2239cc8dff8fbd",
                "depends": [],
                "packages": [[
                    {
                        "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                        'prev': None,
                        'filenames': ["bo_win"],
                        'info': {},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=depwin/0.1 --build=depwin/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            },
            {
                "ref": "depnix/0.1#4d670581ccb765839f2239cc8dff8fbd",
                "depends": [],
                "packages": [[
                    {
                        "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                        'prev': None,
                        'filenames': ["bo_nix"],
                        'info': {},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=depnix/0.1 --build=depnix/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            }
        ],
        [
            {
                "ref": "pkg/0.1#b615ac4c7cd16631cd9e924b68596fce",
                "depends": [
                    "depwin/0.1#4d670581ccb765839f2239cc8dff8fbd",
                    "depnix/0.1#4d670581ccb765839f2239cc8dff8fbd"
                ],
                "packages": [[
                    {
                        "package_id": "b23846b9b10455081d89a9dfacd01f7712d04b95",
                        'prev': None,
                        'filenames': ["bo_win"],
                        'info': {'requires': ['depwin/0.1'],
                                 'settings': {'os': 'Windows'}},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=pkg/0.1 --build=pkg/0.1',
                        "options": [],
                        "overrides": {}
                    },
                    {
                        "package_id": "dc29fa55ec82fab6bd820398c7a152ae5f7d4e28",
                        'prev': None,
                        'filenames': ["bo_nix"],
                        'info': {'requires': ['depnix/0.1'],
                                 'settings': {'os': 'Linux'}},
                        "context": "host",
                        'depends': [],
                        "binary": "Build",
                        'build_args': '--requires=pkg/0.1 --build=pkg/0.1',
                        "options": [],
                        "overrides": {}
                    }
                ]]
            }
        ]
    ]

    assert bo_json == result


def test_info_build_order_lockfile_location():
    """ the lockfile should be in the caller cwd
    https://github.com/conan-io/conan/issues/13850
    """
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/0.1")})
    c.run("create dep")
    c.run("lock create pkg --lockfile-out=myconan.lock")
    assert os.path.exists(os.path.join(c.current_folder, "myconan.lock"))
    c.run("graph build-order pkg --lockfile=myconan.lock --lockfile-out=myconan2.lock")
    assert os.path.exists(os.path.join(c.current_folder, "myconan2.lock"))


def test_build_order_missing_package_check_error():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "pkg/conanfile.py": GenConanfile().with_requires("dep/0.1"),
            "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    c.run("export dep --name=dep --version=0.1")
    c.run("export pkg --name=pkg --version=0.1")

    exit_code = c.run("graph build-order consumer --build='pkg/*' --order=configuration --format=json", assert_error=True)
    bo_json = json.loads(c.stdout)
    assert bo_json["order_by"] == "configuration"
    assert exit_code != 0
    assert "dep/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709: Missing binary" in c.out

    result = [
        [
            {
                "ref": "dep/0.1#4d670581ccb765839f2239cc8dff8fbd",
                "pref": "dep/0.1#4d670581ccb765839f2239cc8dff8fbd:da39a3ee5e6b4b0d3255bfef95601890afd80709",
                "package_id": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                "prev": None,
                "context": "host",
                "binary": "Missing",
                "options": [],
                "filenames": [],
                'info': {},
                "depends": [],
                "overrides": {},
                "build_args": None,
            }
        ],
        [
            {
                "ref": "pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032",
                "pref": "pkg/0.1#1ac8dd17c0f9f420935abd3b6a8fa032:59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                "package_id": "59205ba5b14b8f4ebc216a6c51a89553021e82c1",
                "prev": None,
                "context": "host",
                "binary": "Build",
                "options": [],
                "filenames": [],
                'info': {'requires': ['dep/0.1']},
                "depends": [
                    "dep/0.1#4d670581ccb765839f2239cc8dff8fbd:da39a3ee5e6b4b0d3255bfef95601890afd80709"
                ],
                "overrides": {},
                "build_args": "--requires=pkg/0.1 --build=pkg/0.1",
            }
        ],
    ]

    assert bo_json["order"] == result


def test_info_build_order_broken_recipe():
    # https://github.com/conan-io/conan/issues/14104
    c = TestClient()
    dep = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import replace_in_file
        class Pkg(ConanFile):
            name = "dep"
            version = "0.1"
            def export(self):
                replace_in_file(self, "conanfile.py", "from conan", "from conans")
        """)
    c.save({"conanfile.py": dep})
    c.run("export .")
    c.run("graph build-order --requires=dep/0.1 --format=json", assert_error=True)
    assert "ImportError" in c.out
    assert "It is possible that this recipe is not Conan 2.0 ready" in c.out


class TestBuildOrderReduce:
    @pytest.mark.parametrize("order", ["recipe", "configuration"])
    def test_build_order_reduce(self, order):
        c = TestClient()
        c.save({"liba/conanfile.py": GenConanfile("liba", "0.1"),
                "libb/conanfile.py": GenConanfile("libb", "0.1").with_requires("liba/0.1"),
                "libc/conanfile.py": GenConanfile("libc", "0.1").with_requires("libb/0.1"),
                "consumer/conanfile.txt": "[requires]\nlibc/0.1"})
        c.run("create liba")
        c.run("create libb")
        c.run("create libc")
        c.run("remove liba:* -c")
        c.run("remove libc:* -c")
        c.run(f"graph build-order consumer --order={order} --build=missing --reduce --format=json")
        bo_json = json.loads(c.stdout)
        order_json = bo_json["order"]
        assert len(order_json) == 2  # 2 levels
        level0, level1 = order_json
        assert len(level0) == 1
        assert level0[0]["ref"] == "liba/0.1#a658e7beaaae5d6be0b6f67dcc9859e2"
        # then libc -> directly on liba, no libb involved
        assert len(level1) == 1
        assert level1[0]["ref"] == "libc/0.1#c04c370ad966390e67388565b56f019a"
        depends = "liba/0.1#a658e7beaaae5d6be0b6f67dcc9859e2"
        if order == "configuration":
            depends += ":da39a3ee5e6b4b0d3255bfef95601890afd80709"
        assert level1[0]["depends"] == [depends]

    @pytest.mark.parametrize("order", ["recipe", "configuration"])
    def test_build_order_merge_reduce(self, order):
        c = TestClient()
        c.save({"liba/conanfile.py": GenConanfile("liba", "0.1").with_settings("os"),
                "libb/conanfile.py": GenConanfile("libb", "0.1").with_settings("os")
                                                                .with_requires("liba/0.1"),
                "libc/conanfile.py": GenConanfile("libc", "0.1").with_settings("os")
                                                                .with_requires("libb/0.1"),
                "consumer/conanfile.txt": "[requires]\nlibc/0.1"})
        for _os in ("Windows", "Linux"):
            c.run(f"create liba -s os={_os}")
            c.run(f"create libb -s os={_os}")
            c.run(f"create libc -s os={_os}")

        c.run("remove liba:* -c")
        c.run("remove libc:* -c")
        c.run(f"graph build-order consumer --order={order} --build=missing -s os=Windows "
              "--format=json", redirect_stdout="windows.json")
        c.run(f"graph build-order consumer --order={order} --build=missing -s os=Linux "
              "--format=json", redirect_stdout="linux.json")

        c.run(f"graph build-order-merge --file=windows.json --file=linux.json --reduce "
              "--format=json")
        bo_json = json.loads(c.stdout)
        order_json = bo_json["order"]
        assert len(order_json) == 2  # 2 levels
        level0, level1 = order_json
        if order == "recipe":
            assert len(level0) == 1
            assert level0[0]["ref"] == "liba/0.1#8c6ed89c12ab2ce78b239224bd7cb79e"
            # then libc -> directly on liba, no libb involved
            assert len(level1) == 1
            assert level1[0]["ref"] == "libc/0.1#66db2600b9d6a2a61c9051fcf47da4a3"
            depends = "liba/0.1#8c6ed89c12ab2ce78b239224bd7cb79e"
            assert level1[0]["depends"] == [depends]
        else:
            assert len(level0) == 2
            liba1 = "liba/0.1#8c6ed89c12ab2ce78b239224bd7cb79e:" \
                    "ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715"
            liba2 = "liba/0.1#8c6ed89c12ab2ce78b239224bd7cb79e:" \
                    "9a4eb3c8701508aa9458b1a73d0633783ecc2270"
            assert level0[0]["pref"] == liba1
            assert level0[1]["pref"] == liba2
            # then libc -> directly on liba, no libb involved
            assert len(level1) == 2
            assert level1[0]["ref"] == "libc/0.1#66db2600b9d6a2a61c9051fcf47da4a3"
            assert level1[0]["depends"] == [liba1]
            assert level1[1]["ref"] == "libc/0.1#66db2600b9d6a2a61c9051fcf47da4a3"
            assert level1[1]["depends"] == [liba2]

    def test_error_reduced(self):
        c = TestClient()
        c.save({"conanfile.py": GenConanfile("liba", "0.1")})
        c.run("graph build-order . --format=json", redirect_stdout="bo1.json")
        c.run("graph build-order . --order-by=recipe --reduce --format=json",
              redirect_stdout="bo2.json")
        c.run(f"graph build-order-merge --file=bo1.json --file=bo2.json", assert_error=True)
        assert "ERROR: Reduced build-order file cannot be merged: bo2.json"
        # different order
        c.run(f"graph build-order-merge --file=bo2.json --file=bo1.json", assert_error=True)
        assert "ERROR: Reduced build-order file cannot be merged: bo2.json"

    def test_error_different_orders(self):
        c = TestClient()
        c.save({"conanfile.py": GenConanfile("liba", "0.1")})
        c.run("graph build-order . --format=json", redirect_stdout="bo1.json")
        c.run("graph build-order . --order-by=recipe --format=json", redirect_stdout="bo2.json")
        c.run("graph build-order . --order-by=configuration --format=json",
              redirect_stdout="bo3.json")
        c.run(f"graph build-order-merge --file=bo1.json --file=bo2.json")
        # Not error
        c.run(f"graph build-order-merge --file=bo1.json --file=bo3.json", assert_error=True)
        assert "ERROR: Cannot merge build-orders of recipe!=configuration" in c.out
        c.run(f"graph build-order-merge --file=bo2.json --file=bo3.json", assert_error=True)
        assert "ERROR: Cannot merge build-orders of recipe!=configuration" in c.out
        # different order
        c.run(f"graph build-order-merge --file=bo3.json --file=bo2.json", assert_error=True)
        assert "ERROR: Cannot merge build-orders of configuration!=recipe" in c.out

    def test_merge_missing_error(self):
        tc = TestClient(light=True)
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0")})
        tc.run("export dep")
        tc.run("graph build-order --order=recipe --requires=dep/1.0 --format=json", assert_error=True, redirect_stdout="order.json")
        tc.run("graph build-order-merge --file=order.json --file=order.json --format=json", assert_error=True)
        assert "dep/1.0:da39a3ee5e6b4b0d3255bfef95601890afd80709: Missing binary" in tc.out
        assert "IndexError: list index out of range" not in tc.out

    def test_merge_invalid_error(self):
        tc = TestClient(light=True)
        conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.errors import ConanInvalidConfiguration
        class Pkg(ConanFile):
            name = "dep"
            version = "1.0"
            def validate(self):
                raise ConanInvalidConfiguration("This configuration is not valid")
        """)
        tc.save({"dep/conanfile.py": conanfile})
        tc.run("export dep")
        tc.run("graph build-order --order=recipe --requires=dep/1.0 --format=json", assert_error=True, redirect_stdout="order.json")
        tc.run("graph build-order-merge --file=order.json --file=order.json --format=json", assert_error=True)
        assert "dep/1.0:da39a3ee5e6b4b0d3255bfef95601890afd80709: Invalid configuration" in tc.out
        assert "IndexError: list index out of range" not in tc.out

    def test_reduce_should_remove_recipe(self):
        tc = TestClient()
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0").with_settings("os")})
        tc.run("export dep")
        tc.run("create dep -s os=Windows")
        tc.run("graph build-order -s os=Windows --build=missing --order=recipe --requires=dep/1.0 "
               "--format=json", redirect_stdout="windows.json")
        tc.run("graph build-order -s os=Linux --build=missing --order=recipe --requires=dep/1.0 "
               "--format=json", redirect_stdout="linux.json")
        tc.run("graph build-order-merge --file=windows.json --file=linux.json --reduce "
               "--format=json")
        order = json.loads(tc.stdout)
        assert order["order"][0][0]["ref"] == "dep/1.0#1674c18bb63f0c9778d2811c21f581a0"
        assert len(order["order"][0][0]["packages"][0]) == 1
        assert order["order"][0][0]["packages"][0][0]["binary"] == "Build"


def test_multi_configuration_profile_args():
    c = TestClient()
    c.save({"pkg/conanfile.py": GenConanfile().with_settings("os"),
            "consumer/conanfile.txt": "[requires]\npkg/0.1",
            "mypr": ""})
    c.run("export pkg --name=pkg --version=0.1")
    args = "-pr=mypr -s:b os=Linux -o:h *:shared=True -c:h user.my:conf=1"
    c.run(f"graph build-order consumer --format=json --build=missing -s os=Windows {args} "
          "--order-by=recipe", redirect_stdout="bo_win.json")
    c.run(f"graph build-order consumer --format=json --build=missing -s os=Linux {args} "
          "--order-by=recipe", redirect_stdout="bo_nix.json")
    c.run("graph build-order-merge --file=bo_win.json --file=bo_nix.json --format=json",
          redirect_stdout="bo3.json")
    bo_json = json.loads(c.load("bo3.json"))
    win = '-pr:h="mypr" -s:h="os=Windows" -o:h="*:shared=True" -c:h="user.my:conf=1" -s:b="os=Linux"'
    nix = '-pr:h="mypr" -s:h="os=Linux" -o:h="*:shared=True" -c:h="user.my:conf=1" -s:b="os=Linux"'
    assert bo_json["profiles"] == {"bo_win": {"args": win}, "bo_nix": {"args": nix}}


def test_build_order_space_in_options():
    tc = TestClient(light=True)
    tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0").with_option("flags", ["ANY", None])
                                                           .with_option("extras", ["ANY", None]),
             "conanfile.txt": textwrap.dedent("""
             [requires]
             dep/1.0

             [options]
             dep/*:flags=define=FOO define=BAR define=BAZ
             dep/*:extras=cxx="yes" gnuext='no'
             """)})

    tc.run("create dep")
    tc.run("graph build-order . --order-by=configuration --build=dep/1.0 -f=json", redirect_stdout="order.json")
    order = json.loads(tc.load("order.json"))
    assert order["order"][0][0]["build_args"] == '''--requires=dep/1.0 --build=dep/1.0 -o="dep/*:extras=cxx="yes" gnuext='no'" -o="dep/*:flags=define=FOO define=BAR define=BAZ"'''


def test_build_order_build_context_compatible():
    c = TestClient()
    foo = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.build import check_min_cppstd

        class Pkg(ConanFile):
            name = "foo"
            version = "1.0"
            settings = "os", "compiler"
            def validate_build(self):
                check_min_cppstd(self, 17)

            def validate(self):
                check_min_cppstd(self, 14)
        """)
    bar = GenConanfile("bar", "1.0").with_settings("os", "compiler").with_tool_requirement("foo/1.0")

    profile = textwrap.dedent("""
        [settings]
        compiler=gcc
        compiler.cppstd=gnu14
        compiler.libcxx=libstdc++11
        compiler.version=11
        os=Linux
        """)
    c.save({"conanfile_foo.py": foo,
            "conanfile_bar.py": bar,
            "profile": profile})
    c.run("export conanfile_foo.py")
    c.run("export conanfile_bar.py")

    #  "--require/bar.1.0" and "require=foo/1.0" along with `--build=missing` would cause both
    # packages to be built in the host context - with foo being built with cppstd=17
    # (because the default cppstd=14 is not enough) bar requires foo in the "build" context -
    # where cppstd=14 - but it can reuse the one built for cppstd=17
    # (via compatibility plugin)

    # The three approaches are equivalent:
    #  - Using "--build=missing" and forcing  "-s foo/*:compiler.cppstd=17", means, build missing
    #    binaries, and for foo I want the compiler.cppstd=17 binary. This approach could build other
    #    missing binaries too
    #  - Using "--build=compatible:foo/*" and "--build=missing:bar/*" means, build only missing
    #    binary for bar, and for "foo", build a compatible one if the main one is missing. This
    #    approach prevents other packages (not foo/bar) from accidentally being built.
    #  - Last approach, passing both "--build=missing:foo/*" and "--build=compatible:foo/*" is
    #    similar to the other two in final behavior, but the --build=missing:foo avoids doing the
    #    compatibility check for consumption of "foo", and goes directly to the build check

    for approach in ("--build=missing -s foo/*:compiler.cppstd=17",
                     '--build="compatible:foo/*" --build="missing:bar/*"',
                     '--build="missing:foo/*" --build="compatible:foo/*" --build="missing:bar/*"'):
        c.run(f'graph build-order --require=foo/1.0 --require=bar/1.0 -pr:a profile {approach}')
        c.assert_listed_binary({"foo/1.0": ["4e2ae338231ae18d0d43b9e119404d2b2c416758", "Build"],
                                "bar/1.0": ["5e4ffcc1ff33697a4ee96f66f0d2228ec458f25c", "Build"]})
        c.assert_listed_binary({"foo/1.0": ["4e2ae338231ae18d0d43b9e119404d2b2c416758", "Build"]},
                               build=True)


def test_info_build_order_editable():
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/0.1"),
            "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    c.run("editable add dep")
    c.run("export pkg")

    c.run("graph build-order consumer --build=missing --build=editable -f=json --order-by=recipe")
    bo_json = json.loads(c.stdout)
    pkg = bo_json["order"][0][0]["packages"][0][0]
    assert pkg["binary"] == "EditableBuild"
    assert pkg["build_args"] == "--requires=dep/0.1 --build=dep/0.1"

    c.run("graph build-order consumer --build=missing --build=editable -f=json "
          "--order-by=configuration")
    bo_json = json.loads(c.stdout)
    pkg = bo_json["order"][0][0]
    assert pkg["binary"] == "EditableBuild"
    assert pkg["build_args"] == "--requires=dep/0.1 --build=dep/0.1"


def test_build_order_path_reqs_mixed_args():
    # This used not to crash in previous versions
    # Also make sure we can properly used them separately
    tc = TestClient(light=True)
    tc.run("graph build-order . --requires=foo/1.0", assert_error=True)
    assert "ERROR: --requires and --tool-requires arguments are incompatible with [path] '.' argument" in tc.out

    tc.run("graph build-order --requires=foo/1.0", assert_error=True)
    assert "Package 'foo/1.0' not resolved: No remote defined" in tc.out

    tc.run("graph build-order .", assert_error=True)
    assert "Conanfile not found" in tc.out



import zipfile
import tarfile
from os.path import join, exists

import pytest

from conan.tools.files import unzip
from conan.test.utils.mocks import ConanFileMock
from conan.test.utils.test_files import temp_folder
from conan.internal.util.files import save
from conan.errors import ConanException


def test_impossible_to_import_untargz():
    with pytest.raises(ImportError):
        from conan.tools.files import untargz


def create_example_zip(root_file=True, subfolder=False):
    tmp_dir = temp_folder()
    archive = join(tmp_dir, "zipfile.zip")
    zf = zipfile.ZipFile(archive, mode="w")
    if root_file:
        foo_txt = join(tmp_dir, "foo.txt")
        save(foo_txt, "foo-content")
        zf.write(foo_txt, "foo.txt")
    if subfolder:
        src_bar_txt = join(tmp_dir, "src", "bar.txt")
        save(src_bar_txt, "bar-content")
        zf.write(src_bar_txt, join("src", "bar.txt"))
    zf.close()
    return archive


def test_unzip():
    archive = create_example_zip(subfolder=True)
    conanfile = ConanFileMock({})

    # Unzip and check permissions are kept
    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir)
    assert exists(join(dest_dir, "foo.txt"))
    assert exists(join(dest_dir, "src", "bar.txt"))


def test_unzip_with_pattern():
    archive = create_example_zip(subfolder=True)
    conanfile = ConanFileMock({})

    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir, pattern="foo.txt")
    assert exists(join(dest_dir, "foo.txt"))
    assert not exists(join(dest_dir, "src", "bar.txt"))


def test_unzip_with_exclude_pattern():
    archive = create_example_zip(subfolder=True)
    conanfile = ConanFileMock({})

    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir, excludes=["src/*"])
    assert exists(join(dest_dir, "foo.txt"))
    assert not exists(join(dest_dir, "src", "bar.txt"))


def test_unzip_with_strip_root():
    archive = create_example_zip(root_file=False, subfolder=True)
    conanfile = ConanFileMock({})

    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir, strip_root=True)
    assert exists(join(dest_dir, "bar.txt"))


def test_unzip_with_strip_root_fails():
    archive = create_example_zip(root_file=True, subfolder=True)
    conanfile = ConanFileMock({})

    dest_dir = temp_folder()
    with pytest.raises(ConanException) as error:
        unzip(conanfile, archive, dest_dir, strip_root=True)
    assert "The zip file contains more than 1 folder in the root" in str(error.value)


def test_unzip_with_strip_root_and_pattern():
    archive = create_example_zip(root_file=True, subfolder=True)
    conanfile = ConanFileMock({})

    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir, pattern="src/*", strip_root=True)
    assert exists(join(dest_dir, "bar.txt"))
    assert not exists(join(dest_dir, "foo.txt"))


def create_example_tar(root_file=True, subfolder=False):
    tmp_dir = temp_folder()
    tar_path = join(tmp_dir, "file.tgz")
    tar = tarfile.open(tar_path, "w:gz")
    if root_file:
        foo_txt = join(tmp_dir, "foo.txt")
        save(foo_txt, "foo-content")
        tar.add(foo_txt, "foo.txt")
    if subfolder:
        src_bar_txt = join(tmp_dir, "src", "bar.txt")
        save(src_bar_txt, "bar-content")
        tar.add(src_bar_txt, join("src", "bar.txt"))
    tar.close()
    return tar_path


def test_untargz():
    import io
    from unittest.mock import patch

    archive = create_example_tar(subfolder=True)
    conanfile = ConanFileMock({})

    with patch('sys.stderr', new_callable=io.StringIO) as mock_stderr:
        # Unzip and check permissions are kept
        dest_dir = temp_folder()
        unzip(conanfile, archive, dest_dir)
        assert exists(join(dest_dir, "foo.txt"))
        assert exists(join(dest_dir, "src", "bar.txt"))

        stderr_output = mock_stderr.getvalue()
        assert f"Uncompressing {archive} to {dest_dir}" in stderr_output


def test_untargz_with_pattern():
    archive = create_example_tar(subfolder=True)
    conanfile = ConanFileMock({})

    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir, pattern="foo.txt")
    assert exists(join(dest_dir, "foo.txt"))
    assert not exists(join(dest_dir, "src", "bar.txt"))


def test_untargz_with_exclude_pattern():
    archive = create_example_tar(subfolder=True)
    conanfile = ConanFileMock({})

    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir, excludes=["src/*"])
    assert exists(join(dest_dir, "foo.txt"))
    assert not exists(join(dest_dir, "src", "bar.txt"))


def test_untargz_with_strip_root():
    archive = create_example_tar(root_file=False, subfolder=True)
    conanfile = ConanFileMock({})

    # Unzip and check permissions are kept
    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir, strip_root=True)
    assert exists(join(dest_dir, "bar.txt"))


def test_untargz_with_strip_root_fails():
    archive = create_example_tar(root_file=True, subfolder=True)
    conanfile = ConanFileMock({})

    # Unzip and check permissions are kept
    dest_dir = temp_folder()
    with pytest.raises(ConanException) as error:
        unzip(conanfile, archive, dest_dir, strip_root=True)
    assert "Can't untar a tgz containing files in the root with strip_root enabled" in str(error.value)


def test_untargz_with_strip_root_and_pattern():
    archive = create_example_tar(root_file=True, subfolder=True)
    conanfile = ConanFileMock({})

    # Unzip and check permissions are kept
    dest_dir = temp_folder()
    unzip(conanfile, archive, dest_dir, pattern="src/*", strip_root=True)
    assert exists(join(dest_dir, "bar.txt"))
    assert not exists(join(dest_dir, "foo.txt"))



import pytest

from conan.internal.model.version import Version

values = [
    ['1.0.0',       0, "2.0.0"],
    ['1.1.0',       0, "2.0.0"],
    ['1.1.1-pre',   0, "2.0.0"],
    ['1.1.1',       1, "1.2.0"],
    ['1.1.1',       2, "1.1.2"],
]


@pytest.mark.parametrize("version, index, result", values)
def test_version_bump(version, index, result):
    r = Version(version)
    bumped = r.bump(index)
    assert bumped == result



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


class TestTransitiveOptionsAffectPackageID:

    def test_basic(self):
        client = TestClient()
        conanfile = GenConanfile("pkg", "0.1").with_option("noaffect", [1, 2])\
                                              .with_option("affect", [True, False])\
                                              .with_package_type("static-library")
        client.save({"conanfile.py": conanfile})
        client.run("create . -o pkg*:noaffect=1 -o pkg*:affect=False")
        client.assert_listed_binary({"pkg": ("7ff068ae587920b4f40b0dd81e891808c419f78c", "Build")})

        client.run("create . -o pkg*:noaffect=2 -o pkg*:affect=False")
        client.assert_listed_binary({"pkg": ("66de80286b3d7ae1d5c05a9b795d6a6e622a778e", "Build")})

        client.run("create . -o pkg*:noaffect=1 -o pkg*:affect=True")
        client.assert_listed_binary({"pkg": ("8e0cdf1d5b1c1c557b6f4b4ec01a4383b8f50dc3", "Build")})

        client.run("create . -o pkg*:noaffect=2 -o pkg*:affect=True")
        client.assert_listed_binary({"pkg": ("4a18825b7c66155f03b27a1963e42ab3a836fd28", "Build")})

        app = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "app"
                version = "0.1"
                package_type = "application"
                requires = "pkg/0.1"
                def package_id(self):
                    if self.dependencies["pkg"].package_type == "static-library":
                        self.output.info("EMBED MODE!!")
                        self.info.requires["pkg"].package_id = None
                        self.info.options["pkg/*"].affect = self.dependencies["pkg"].options.affect
            """)
        client.save({"conanfile.py": app})
        client.run("create . -o pkg*:noaffect=1 -o pkg*:affect=False")
        assert "EMBED MODE!!" in client.out
        common_pkg_id = "2060adfe3cc9040e5ee7e54c783d752f9ca1ba43"  # Should be the same!
        client.assert_listed_binary({"app": (common_pkg_id, "Build")})
        client.run("create . -o pkg*:noaffect=2 -o pkg*:affect=False")
        assert "EMBED MODE!!" in client.out
        client.assert_listed_binary({"app": (common_pkg_id, "Build")})

        client.run("create . -o pkg*:noaffect=1 -o pkg*:affect=True")
        assert "EMBED MODE!!" in client.out
        different_pkg_id = "acfad205713f0d2fce61e7927d059635c0abbfff"  # Should be the same!
        client.assert_listed_binary({"app": (different_pkg_id, "Build")})
        client.run("create . -o pkg*:noaffect=2 -o pkg*:affect=True")
        assert "EMBED MODE!!" in client.out
        client.assert_listed_binary({"app": (different_pkg_id, "Build")})

    def test_transitive_shared(self):
        # https://github.com/conan-io/conan/issues/18900
        c = TestClient()

        lib1 = GenConanfile("lib1", "0.1").with_shared_option(True)
        lib2 = (GenConanfile("lib2", "0.1").with_shared_option(True)
                .with_requirement("lib1/0.1", transitive_libs=True))
        lib3 = (GenConanfile("lib3", "0.1").with_shared_option(True)
                .with_requirement("lib2/0.1"))
        lib4 = (GenConanfile("lib4", "0.1").with_shared_option(True)
                .with_requirement("lib3/0.1"))
        c.save({"lib1/conanfile.py": lib1,
                "lib2/conanfile.py": lib2,
                "lib3/conanfile.py": lib3,
                "lib4/conanfile.py": lib4})
        c.run("create lib1")
        c.run("create lib2")
        c.run("create lib3")
        c.run("create lib4")

        c.run("install --requires=lib4/0.1 -o lib1/*:shared=False --build=missing")
        c.assert_listed_binary({"lib1/0.1": ("55c609fe8808aa5308134cb5989d23d3caffccf2", "Build"),
                                "lib2/0.1": ("76844632e497abea8503d65ffd8324460dc70745", "Build"),
                                "lib3/0.1": ("7b10301e532fc0269d6ac70470aee5780f0836cd", "Build"),
                                "lib4/0.1": ("635372f179ad582c713637e361a7cd7ac7cd1d09", "Cache"),
                                })

        lib3 = (GenConanfile("lib3", "0.1").with_shared_option(True)
                .with_requirement("lib2/0.1", transitive_libs=True))
        c.save({"lib3/conanfile.py": lib3})
        c.run("remove * -c")
        c.run("create lib1")
        c.run("create lib2")
        c.run("create lib3")
        c.run("create lib4")
        c.run("install --requires=lib4/0.1 -o lib1/*:shared=False --build=missing")
        c.assert_listed_binary({"lib1/0.1": ("55c609fe8808aa5308134cb5989d23d3caffccf2", "Build"),
                                "lib2/0.1": ("76844632e497abea8503d65ffd8324460dc70745", "Build"),
                                "lib3/0.1": ("7b10301e532fc0269d6ac70470aee5780f0836cd", "Build"),
                                "lib4/0.1": ("dd8f5355b399fd7d96c883ddd39b992ae968cb14", "Build"),
                                })



import re

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.test.utils.env import environment_update


@pytest.mark.artifactory_ready
class TestDownloadPatterns:
    # The fixture is very similar from TestUploadPatterns, but not worth extracting
    @pytest.fixture(scope="class")
    def client(self):
        """ create a few packages, with several recipe revisions, several pids, several prevs
        """
        client = TestClient(default_server_user=True)

        for pkg in ("pkga", "pkgb"):
            for version in ("1.0", "1.1"):
                for rrev in ("rev1", "rev2"):
                    client.save({"conanfile.py": GenConanfile().with_settings("os")
                                .with_class_attribute(f"potato='{rrev}'")
                                .with_package_file("file", env_var="MYVAR")})
                    for the_os in ("Windows", "Linux"):
                        for prev in ("prev1", "prev2"):
                            with environment_update({"MYVAR": prev}):
                                client.run(f"create . --name={pkg} --version={version} "
                                           f"-s os={the_os}")
        client.run("upload *#*:*#* -r=default -c")
        return client

    @staticmethod
    def assert_downloaded(pattern, result, client, only_recipe=False, query=None):
        def ref_map(r):
            rev1 = "ad55a66b62acb63ffa99ea9b75c16b99"
            rev2 = "127fb537a658ad6a57153a038960dc53"
            pid1 = "ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715"
            pid2 = "9a4eb3c8701508aa9458b1a73d0633783ecc2270"
            if "Linux" in r:
                prev1 = "7ce684c6109943482b9174dd089e717b"
                prev2 = "772234192e8e4ba71b018d2e7c02423e"
            else:
                prev1 = "45f88f3c318bd43d1bc48a5d408a57ef"
                prev2 = "c5375d1f517ecb1ed6c9532b0f4d86aa"

            r = r.replace("prev1", prev1).replace("prev2", prev2).replace("Windows", pid1)
            r = r.replace("Linux", pid2).replace("rev1", rev1).replace("rev2", rev2)
            return r

        pattern = ref_map(pattern)
        only_recipe = "" if not only_recipe else "--only-recipe"
        query = "" if not query else f"-p={query}"
        client.run(f"download {pattern} -r=default {only_recipe} {query}")
        out = str(client.out)

        downloaded_recipes = [f"{p}/{v}#{rr}" for p in result[0]
                              for v in result[1]
                              for rr in result[2]]
        downloaded_packages = [f"{r}:{pid}#{pr}" for r in downloaded_recipes
                               for pid in result[3]
                               for pr in result[4]]

        # Checks
        skipped_recipe_count = len(re.findall("Skip recipe .+ download, already in cache", out))
        assert skipped_recipe_count == len(downloaded_recipes)
        for recipe in downloaded_recipes:
            recipe = ref_map(recipe)
            existing = f"Skip recipe {recipe} download, already in cache" in out
            assert existing
        skipped_pkg_count = len(re.findall("Skip package .+ download, already in cache", out))
        assert skipped_pkg_count == len(downloaded_packages)
        for pkg in downloaded_packages:
            pkg = ref_map(pkg)
            existing = f"Skip package {pkg} download, already in cache" in out
            assert existing

    def test_all_latest(self, client):
        result = ("pkga", "pkgb"), ("1.0", "1.1"), ("rev2",), ("Windows", "Linux"), ("prev2",)
        self.assert_downloaded("*", result, client)

    def test_all(self, client):
        result = ("pkga", "pkgb"), ("1.0", "1.1"), ("rev1", "rev2",), ("Windows", "Linux"), \
                 ("prev1", "prev2")
        self.assert_downloaded("*#*:*#*", result, client)

    def test_pkg(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev2",), ("Windows", "Linux"), ("prev2",)
        self.assert_downloaded("pkga", result, client)

    def test_pkg_rrev(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev1",), ("Windows", "Linux"), ("prev2",)
        self.assert_downloaded("pkga#rev1", result, client)

    def test_pkg_rrevs(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev1", "rev2"), ("Windows", "Linux"), ("prev2",)
        self.assert_downloaded("pkga#*", result, client)

    def test_pkg_pid(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev2",), ("Windows",), ("prev2",)
        self.assert_downloaded("pkga:Windows", result, client)

    def test_pkg_rrev_pid(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev1",), ("Windows",), ("prev2",)
        self.assert_downloaded("pkga#rev1:Windows", result, client)

    def test_pkg_rrevs_pid(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev1", "rev2"), ("Windows",), ("prev2",)
        self.assert_downloaded("pkga#*:Windows", result, client)

    def test_pkg_rrev_pid_prev(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev1",), ("Windows",), ("prev1",)
        self.assert_downloaded("pkga#rev1:Windows#prev1", result, client)

    # Only recipes
    def test_all_latest_only_recipe(self, client):
        result = ("pkga", "pkgb"), ("1.0", "1.1"), ("rev2",), (), ()
        self.assert_downloaded("*", result, client, only_recipe=True)

    def test_pkg_only_recipe(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev2",), (), ()
        self.assert_downloaded("pkga", result, client, only_recipe=True)

    def test_pkg_rrev_only_recipe(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev1",), (), ()
        self.assert_downloaded("pkga#rev1", result, client, only_recipe=True)

    def test_pkg_rrevs_only_recipe(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev1", "rev2"), (), ()
        self.assert_downloaded("pkga#*", result, client, only_recipe=True)

    # Package query
    def test_all_query(self, client):
        result = ("pkga", "pkgb"), ("1.0", "1.1"), ("rev2",), ("Windows",), ("prev2",)
        self.assert_downloaded("*", result, client, query="os=Windows")

    def test_pkg_query(self, client):
        result = ("pkga",), ("1.0", "1.1"), ("rev2",), ("Windows",), ("prev2",)
        self.assert_downloaded("pkga", result, client, query="os=Windows")


class TestDownloadPatterErrors:

    @pytest.fixture(scope="class")
    def client(self):
        client = TestClient(default_server_user=True)
        client.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        client.run(f"create .")
        client.run("upload *#*:*#* -r=default -c")
        return client

    @staticmethod
    def assert_error(pattern, error, client, only_recipe=False, query=None):
        only_recipe = "" if not only_recipe else "--only-recipe"
        query = "" if not query else f"-p={query}"
        client.run(f"download {pattern} -r=default {only_recipe} {query}", assert_error=True)
        assert error in client.out

    def test_recipe_not_found(self, client):
        # FIXME: This error is ugly
        error = "ERROR: Recipe not found: 'zlib/1.2.11@_/_'. [Remote: default]"
        self.assert_error("zlib/1.2.11", error, client)

    def test_rrev_not_found(self, client):
        error = "ERROR: Recipe revision 'pkg/0.1#rev1' not found"
        self.assert_error("pkg/0.1#rev1", error, client)

    def test_pid_not_found(self, client):
        # FIXME: Ugly error, improve it
        rrev = "485dad6cb11e2fa99d9afbe44a57a164"
        error = f"Binary package not found: 'pkg/0.1@_/_#{rrev}:pid1'. [Remote: default]"
        self.assert_error(f"pkg/0.1#{rrev}:pid1", error, client)

    def test_prev_not_found(self, client):
        rrev = "485dad6cb11e2fa99d9afbe44a57a164"
        pid = "da39a3ee5e6b4b0d3255bfef95601890afd80709"
        error = f"ERROR: Package revision 'pkg/0.1#{rrev}:{pid}#prev' not found"
        self.assert_error(f"pkg/0.1#{rrev}:{pid}#prev", error, client)



import os
import textwrap

import pytest

from conan.internal.model.manifest import FileTreeManifest
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.parametrize("relative_path", [False, True])
def test_basic(relative_path):
    client = TestClient()
    source_folder = os.path.join(client.current_folder, "source")
    files = {"conanfile.py": GenConanfile().with_exports("*"),
             "main.cpp": "mymain"}
    client.save(files, path=source_folder)
    if relative_path:
        with client.chdir("current"):
            client.run("export ../source --name=hello --version=0.1 --user=lasote --channel=stable")
    else:
        client.run("export source --name=hello --version=0.1 --user=lasote --channel=stable")

    # The result should be the same in both cases
    ref_layoyt = client.exported_layout()
    ref = ref_layoyt.reference
    reg_path = ref_layoyt.export()
    manif = FileTreeManifest.load(reg_path)

    assert '%s: Exported' % str(ref) in client.out
    assert '%s: Exported to cache folder: %s' % (str(ref), reg_path) in client.out

    for name in list(files.keys()):
        assert os.path.exists(os.path.join(reg_path, name))

    expected_sums = {'conanfile.py': '7fbb7e71f5b665780ff8c407fe0b1f9e',
                     'main.cpp': '76c0a7a9d385266e27d69d3875f6ac19'}
    assert expected_sums == manif.file_sums


@pytest.mark.parametrize("relative_path", [False, True])
def test_path(relative_path):
    client = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy
        class Pkg(ConanFile):
            def export(self):
                copy(self, "*", src=os.path.join(self.recipe_folder, "..", "source"),
                     dst=os.path.join(self.export_folder, "source"))
            """)
    if relative_path:
        client.save({"conan/conanfile.py": conanfile,
                     "source/main.cpp": "mymain"})
        with client.chdir("current"):
            client.run("export ../conan --name=hello --version=0.1 --user=lasote --channel=stable")
    else:
        client.save({"current/conanfile.py": conanfile,
                     "source/main.cpp": "mymain"})
        with client.chdir("current"):
            client.run("export --name=hello --version=0.1 --user=lasote --channel=stable")
    ref_layoyt = client.exported_layout()
    reg_path = ref_layoyt.export()
    manif = FileTreeManifest.load(reg_path)

    for name in ['conanfile.py', 'conanmanifest.txt', 'source/main.cpp']:
        assert os.path.exists(os.path.join(reg_path, name))

    expected_sums = {'conanfile.py': '6cdb33126a0408bffc0ad0ada66cb061',
                     'source/main.cpp': '76c0a7a9d385266e27d69d3875f6ac19'}
    assert expected_sums == manif.file_sums



import os

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.scm import git_add_changes_commit
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


@pytest.mark.tool("git")
class TestRevisionModeSCM:

    def test_revision_mode_scm(self):
        t = TestClient()
        conanfile = str(GenConanfile().with_class_attribute('revision_mode = "scm"'))
        commit = t.init_git_repo({'conanfile.py': conanfile})

        t.run(f"export . --name=pkg --version=0.1")
        assert t.exported_recipe_revision() == commit

        # Now it will fail if dirty
        t.save({"conanfile.py": conanfile + "\n#comment"})
        t.run(f"export . --name=pkg --version=0.1", assert_error=True)
        assert "Can't have a dirty repository using revision_mode='scm' and doing" in t.out
        # Commit to fix
        commit2 = git_add_changes_commit(t.current_folder, msg="fix")
        t.run(f"export . --name=pkg --version=0.1")
        assert t.exported_recipe_revision() == commit2

    def test_revision_mode_scm_subfolder(self):
        """ emulates a mono-repo with 2 subprojects, when a change is done in a subproject
        it gets a different folder commit
        """
        t = TestClient()
        conanfile = str(GenConanfile().with_class_attribute('revision_mode = "scm_folder"'))
        commit = t.init_git_repo({'pkga/conanfile.py': conanfile,
                                  'pkgb/conanfile.py': conanfile})

        t.save({"pkgb/conanfile.py": conanfile + "\n#comment"})
        commit_b = git_add_changes_commit(os.path.join(t.current_folder, "pkgb"), msg="fix")

        # pkga still gets the initial commit, as it didn't change its contents
        t.run(f"export pkga --name=pkga --version=0.1")
        assert t.exported_recipe_revision() == commit

        # but pkgb will get the commit of the new changed folder
        t.run(f"export pkgb --name=pkgb --version=0.1")
        assert t.exported_recipe_revision() == commit_b

        # if pkgb is dirty, we should still be able to correctly create pkga in 'scm_folder' mode
        t.save({"pkgb/conanfile.py": conanfile + "\n#new comment"})
        t.run(f"export pkga --name=pkga --version=0.1")
        assert t.exported_recipe_revision() == commit

    def test_auto_revision_without_commits(self):
        """If we have a repo but without commits, it has to fail when the revision_mode=scm"""
        t = TestClient()
        t.run_command('git init .')
        t.save({"conanfile.py": GenConanfile("lib", "0.1").with_revision_mode("scm")})
        t.run("export .", assert_error=True)
        # It errors, because no commits yet
        assert "Cannot detect revision using 'scm' mode from repository" in t.out

    @pytest.mark.parametrize("conf_excluded, recipe_excluded",
                             [("", ["*.cpp", "*.txt", "src/*"]),
                              (["*.cpp", "*.txt", "src/*"], ""),
                              ('+["*.cpp", "*.txt"]', ["src/*"]),
                              ('+["*.cpp"]', ["*.txt", "src/*"])])
    def test_revision_mode_scm_excluded_files(self, conf_excluded, recipe_excluded):
        t = TestClient()
        recipe_excluded = f'revision_mode_excluded = {recipe_excluded}' if recipe_excluded else ""
        conf_excluded = f'core.scm:excluded={conf_excluded}' if conf_excluded else ""
        save(t.paths.global_conf_path, conf_excluded)
        conanfile = GenConanfile("pkg", "0.1").with_class_attribute('revision_mode = "scm"') \
                                              .with_class_attribute(recipe_excluded)
        commit = t.init_git_repo({'conanfile.py': str(conanfile),
                                  "test.cpp": "mytest"})

        t.run(f"export .")
        assert t.exported_recipe_revision() == commit

        t.save({"test.cpp": "mytest2",
                "new.txt": "new",
                "src/potato": "hello"})
        t.run(f"export . -vvv")
        assert t.exported_recipe_revision() == commit

        t.save({"test.py": ""})
        t.run(f"export .", assert_error=True)
        assert "ERROR: Can't have a dirty repository using revision_mode='scm'" in t.out
